@@ -3,7 +3,7 @@
 /*  bam2bcf.c -- variant calling.
 
     Copyright (C) 2010-2012 Broad Institute.
-    Copyright (C) 2012-2021 Genome Research Ltd.
+    Copyright (C) 2012-2024 Genome Research Ltd.
 
     Author: Heng Li <lh3@sanger.ac.uk>
 
@@ -74,9 +74,11 @@ bcf_callaux_t *bcf_call_init(double theta, int min_baseQ, int max_baseQ,
     return bca;
 }
 
+void bcf_iaux_destroy(bcf_callaux_t *bca);
 void bcf_call_destroy(bcf_callaux_t *bca)
 {
     if (bca == 0) return;
+    bcf_iaux_destroy(bca);
     errmod_destroy(bca->e);
     if (bca->npos) {
         free(bca->ref_pos);  free(bca->alt_pos);
@@ -91,6 +93,55 @@ void bcf_call_destroy(bcf_callaux_t *bca)
     free(bca->bases); free(bca->inscns); free(bca);
 }
 
+static int get_aux_nm(const bam_pileup1_t *p, int32_t qpos, int is_ref)
+{
+    int64_t nm;
+    const bam_pileup_cd *cd = &p->cd;
+
+    if ( PLP_NM(cd) == -1 ) return -1;
+    if ( PLP_NM(cd) == PLP_NM_UNSET )
+    {
+        // todo: make this localized to be useful for long reads as well
+        bam1_t *rec = p->b;
+        uint8_t *nm_tag = bam_aux_get(rec, "NM");
+        if ( !nm_tag )
+        {
+            PLP_NM(cd) = -1;
+            return -1;
+        }
+        nm = bam_aux2i(nm_tag);
+
+        // Count indels as single events, not as the number of inserted/deleted
+        // bases (which is what NM does). Add soft clips as mismatches.
+        int i;
+        for (i=0; i < rec->core.n_cigar; i++)
+        {
+            int val = bam_get_cigar(rec)[i] & BAM_CIGAR_MASK;
+            if ( val==BAM_CSOFT_CLIP )
+            {
+                nm += bam_get_cigar(rec)[i] >> BAM_CIGAR_SHIFT;
+            }
+            else if ( val==BAM_CINS || val==BAM_CDEL )
+            {
+                val = bam_get_cigar(rec)[i] >> BAM_CIGAR_SHIFT;
+                if ( val > 1 ) nm -= val - 1;
+            }
+        }
+        PLP_NM(cd) = nm;
+    }
+    else
+        nm = PLP_NM(cd);
+
+    // Take into account MNPs, 2% of de novo SNVs appear within 20bp of another de novo SNV
+    //      http://www.genome.org/cgi/doi/10.1101/gr.239756.118
+    nm -= is_ref ? 1 : 2;
+
+    if ( nm < 0 ) nm = 0;
+    if ( nm >= B2B_N_NM ) nm = B2B_N_NM - 1;
+
+    return nm;
+}
+
 // position in the sequence with respect to the aligned part of the read
 static int get_position(const bam_pileup1_t *p, int *len,
                         int *sc_len, int *sc_dist) {
@@ -160,11 +211,25 @@ void bcf_callaux_clean(bcf_callaux_t *bca, bcf_call_t *call)
     if ( call->ADF ) memset(call->ADF,0,sizeof(int32_t)*(call->n+1)*B2B_MAX_ALLELES);
     if ( call->ADR ) memset(call->ADR,0,sizeof(int32_t)*(call->n+1)*B2B_MAX_ALLELES);
     if ( call->SCR ) memset(call->SCR,0,sizeof(*call->SCR)*(call->n+1));
+    if ( call->SCR ) memset(call->SCR,0,sizeof(*call->SCR)*(call->n+1));
+    if ( bca->fmt_flag&B2B_FMT_NMBZ )
+    {
+        memset(call->ref_nm,0,sizeof(*call->ref_nm)*(call->n+1)*B2B_N_NM);
+        memset(call->alt_nm,0,sizeof(*call->alt_nm)*(call->n+1)*B2B_N_NM);
+    }
+    else
+    {
+        memset(call->ref_nm,0,sizeof(*call->ref_nm)*B2B_N_NM);
+        memset(call->alt_nm,0,sizeof(*call->alt_nm)*B2B_N_NM);
+    }
     memset(call->QS,0,sizeof(*call->QS)*call->n*B2B_MAX_ALLELES);
     memset(bca->ref_scl,  0, 100*sizeof(int));
     memset(bca->alt_scl,  0, 100*sizeof(int));
     memset(bca->iref_scl, 0, 100*sizeof(int));
     memset(bca->ialt_scl, 0, 100*sizeof(int));
+    int i;
+    for (i=0; i<2; i++) bca->nnm[i] = 0;
+    for (i=0; i<2; i++) bca->nm[i] = 0;
 }
 
 /*
@@ -186,6 +251,10 @@ int bcf_call_glfgen(int _n, const bam_pileup1_t *pl, int ref_base, bcf_callaux_t
 {
     int i, n, ref4, is_indel, ori_depth = 0;
 
+#ifdef GLF_DEBUG
+    fprintf(bcftools_stderr, "Call GLFGEN\n");
+#endif
+
     // clean from previous run
     r->ori_depth = 0;
     r->mq0 = 0;
@@ -205,24 +274,96 @@ int bcf_call_glfgen(int _n, const bam_pileup1_t *pl, int ref_base, bcf_callaux_t
         bca->bases = (uint16_t*)realloc(bca->bases, 2 * bca->max_bases);
     }
 
+    // Detect if indel occurs anywhere in this sample
+    int indel_in_sample = 0;
+    if (bca->edlib) {
+        for (i = n = 0; i < _n; ++i) {
+            const bam_pileup1_t *p = pl + i;
+            if (p->indel) indel_in_sample = 1;
+        }
+    }
+
     // fill the bases array
     double nqual_over_60 = bca->nqual / 60.0;
     int ADR_ref_missed[4] = {0};
     int ADF_ref_missed[4] = {0};
     for (i = n = 0; i < _n; ++i) {
         const bam_pileup1_t *p = pl + i;
-        int q, b, mapQ, baseQ, is_diff, min_dist, seqQ;
-        if ( bca->fmt_flag&(B2B_INFO_SCR|B2B_FMT_SCR) && PLP_HAS_SOFT_CLIP(p->cd.i) ) r->SCR++;
+        int b;          // the base or indel type
+        int q;          // the base or indel quality used to calculate PL
+        int seqQ;       // used to cap the indel quality given the sequence context
+        int mapQ;       // to cap the quality for low MQ reads
+        int baseQ;      // used only for supporting INFO annotations
+        int is_diff;    // is this base or indel type different from the reference
+        int min_dist;   // distance from the end, used for tail distance bias
+        if ( bca->fmt_flag&(B2B_INFO_SCR|B2B_FMT_SCR) && PLP_HAS_SOFT_CLIP(&p->cd) ) r->SCR++;
         if (p->is_refskip || (p->b->core.flag&BAM_FUNMAP)) continue;
-        if (p->is_del && !is_indel) continue;
+
+        // The meaning of the indel related variables:
+        //  is_indel  .. is this position currently tested for an indel
+        //  p->is_del .. is the current base a deletion in this read (unrelated to the tested indel)
+        //  p->indel  .. is there an indel starting after this position (i.e. does this read have the tested indel)
+        if (p->is_del && !is_indel) continue;   // not testing an indel and the read has a spanning deletion
+
+        int inm = -1;
+
         ++ori_depth;
-        if (is_indel)
+        if (is_indel)   // testing an indel position
         {
-            b = p->aux>>16&0x3f;
+            b = p->aux>>16&0x3f;        // indel type
             seqQ = q = (p->aux & 0xff); // mp2 + builtin indel-bias
+
+            if (bca->edlib) {
+                if (indel_in_sample) {
+                    seqQ = q = (p->aux & 0xff); // mp2 + builtin indel-bias
+                } else if (p->aux & 0xff) {
+                    // An indel in another sample, but not this.  So just use
+                    // basic sequence confidences.
+                    q = bam_get_qual(p->b)[p->qpos];
+                    if (q > bca->max_baseQ) q = bca->max_baseQ;
+                    seqQ = 99;
+                }
+            }
+
+            if ( !bca->indels_v20 && !bca->edlib )
+            {
+                /*
+                    This heuristics was introduced by e4e161068 and claims to fix #1446. However, we obtain
+                    correct result on the provided test case even when this code is commented out, so this
+                    may not be needed anymore. Leaving it in only for backward compatibility for now.
+                    See mpileup-tests homdel-issue-1446 and CHM1_CHM13_2.45x-1-1701408 which work only when
+                    this code is disabled.
+                */
+                if (p->indel == 0 && (q < _n/2 || _n > 20)) {
+                    // high quality indel calls without p->indel set aren't
+                    // particularly indicative of being a good REF match either,
+                    // at least not in low coverage.  So require solid coverage
+                    // before we start utilising such quals.
+                    b = 0;
+                    q = (int)bam_get_qual(p->b)[p->qpos];
+                    seqQ = (3*seqQ + 2*q)/8;
+                }
+                if (_n > 20 && seqQ > 40) seqQ = 40;
+            }
+
+            is_diff = b ? 1 : 0;
+            if ( bca->fmt_flag&(B2B_FMT_NMBZ|B2B_INFO_NMBZ|B2B_INFO_NM) )
+            {
+                inm = get_aux_nm(p,p->qpos,is_diff?0:1);
+                if ( inm>=0 )
+                {
+                    bca->nnm[is_diff]++;
+                    bca->nm[is_diff] += inm;
+                }
+            }
+
+#ifdef GLF_DEBUG
+            fprintf(bcftools_stderr, "GLF %s\t%d\t%d\n", bam_get_qname(p->b),
+                    bca->indel_types[b], q);
+#endif
             if (q < bca->min_baseQ)
             {
-                if (!p->indel && b < 4)
+                if (!p->indel && b < 4) // not an indel read
                 {
                     if (bam_is_rev(p->b))
                         ADR_ref_missed[b]++;
@@ -231,19 +372,51 @@ int bcf_call_glfgen(int _n, const bam_pileup1_t *pl, int ref_base, bcf_callaux_t
                 }
                 continue;
             }
-            if (p->indel == 0 && (q < _n/2 || _n > 20)) {
-                // high quality indel calls without p->indel set aren't
-                // particularly indicative of being a good REF match either,
-                // at least not in low coverage.  So require solid coverage
-                // before we start utilising such quals.
-                b = 0;
-                q = (int)bam_get_qual(p->b)[p->qpos];
-                seqQ = (3*seqQ + 2*q)/8;
-            }
-            if (_n > 20 && seqQ > 40) seqQ = 40;
-            baseQ  = p->aux>>8&0xff;
 
-            is_diff = (b != 0);
+#ifndef MIN
+#define MIN(a,b) ((a)<(b)?(a):(b))
+#endif
+
+#ifndef MAX
+#define MAX(a,b) ((a)>(b)?(a):(b))
+#endif
+
+#if 1 // TEST 6
+            if (bca->edlib) {
+                // Deeper data should rely more heavily on counts of data
+                // than quality, as quality can be unreliable and prone to
+                // miscalculations through BAQ, STR analysis, etc.
+                // So we put a cap on how good seqQ can be.
+                //
+                // Is it simply the equivalent of increasing -F filter?
+                // Not quite, as the latter removes many real variants upfront.
+                // This calls them and then post-adjusts quality, potentially
+                // dropping it later or changing genotype. So we still get
+                // calls, but lower qual.
+                seqQ = MIN(seqQ, bca->seqQ_offset-(MIN(20,_n)*5));
+
+                if (indel_in_sample && p->indel == 0 && b != 0) {
+                    // This read doesn't contain an indel in CIGAR, but it
+                    // is assigned to an indel now (b != 0),  These are
+                    // reads we've corrected with realignment, but they're
+                    // also enriched for FPs so at high depth we reduce their
+                    // confidence and let the depth do the talking.  If it's
+                    // real and deep, then we don't need every read aligning.
+                    // We also reduce base quality too to reflect the
+                    // chance of our realignment being incorrect.
+
+                    seqQ = MIN(seqQ, seqQ/2 + 5); // q2p5
+
+                    // Finally reduce indel quality.
+                    // This is a blend of indelQ and base QUAL.
+                    q = MIN((int)bam_get_qual(p->b)[p->qpos]/4+10, q/4+1);
+                }
+            }
+#endif
+
+            // Note baseQ changes some output fields such as I16, but has no
+            // significant affect on "call".
+            baseQ  = p->aux>>8&0xff;
         }
         else
         {
@@ -265,15 +438,31 @@ int bcf_call_glfgen(int _n, const bam_pileup1_t *pl, int ref_base, bcf_callaux_t
             baseQ = q;
             seqQ  = 99;
             is_diff = (ref4 < 4 && b == ref4)? 0 : 1;
+            if ( bca->fmt_flag&(B2B_FMT_NMBZ|B2B_INFO_NMBZ|B2B_INFO_NM) )
+            {
+                inm = get_aux_nm(p,p->qpos,is_diff?0:1);
+                if ( inm>=0 )
+                {
+                    bca->nnm[is_diff]++;
+                    bca->nm[is_diff] += inm;
+                }
+            }
         }
         mapQ  = p->b->core.qual < 255? p->b->core.qual : DEF_MAPQ; // special case for mapQ==255
         if ( !mapQ ) r->mq0++;
+#ifdef GLF_DEBUG
+        fprintf(bcftools_stderr, "GLF2 %s\t%d\t%d\t%d,%d\n",
+                bam_get_qname(p->b), b, q,
+                seqQ, mapQ);
+#endif
         if (q > seqQ) q = seqQ;
         mapQ = mapQ < bca->capQ? mapQ : bca->capQ;
         if (q > mapQ) q = mapQ;
         if (q > 63) q = 63;
         if (q < 4) q = 4;       // MQ=0 reads count as BQ=4
         bca->bases[n++] = q<<5 | (int)bam_is_rev(p->b)<<4 | b;
+        //if (is_indel) fprintf(bcftools_stderr,"xx:base,q,strand\t%d\t%d\t%d\n",b,q,bam_is_rev(p->b)?0:1);
+
         // collect annotations
         if (b < 4)
         {
@@ -301,17 +490,17 @@ int bcf_call_glfgen(int _n, const bam_pileup1_t *pl, int ref_base, bcf_callaux_t
         if ( baseQ > 59 ) baseQ = 59;
         if ( mapQ > 59 ) mapQ = 59;
         int len, epos = 0, sc_len = 0, sc_dist = 0;
-        if ( bca->fmt_flag & (B2B_INFO_RPB|B2B_INFO_VDB|B2B_INFO_SCB) )
+        if ( bca->fmt_flag & (B2B_INFO_RPBZ|B2B_INFO_VDB|B2B_INFO_SCBZ) )
         {
             int pos = get_position(p, &len, &sc_len, &sc_dist);
-            epos = (double)pos/(len+1) * bca->npos;
-
+            epos = (double)pos/(len+1) * (bca->npos - 1);
             if (sc_len) {
-                sc_len = 15.0*sc_len / sc_dist;
+                sc_len = 15.0*sc_len / (sc_dist+1);
                 if (sc_len > 99) sc_len = 99;
             }
+            assert( epos>=0 && epos<bca->npos );
+            assert( sc_len>=0 && sc_len<bca->npos );
         }
-
         int imq  = mapQ * nqual_over_60;
         int ibq  = baseQ * nqual_over_60;
 
@@ -320,12 +509,17 @@ int bcf_call_glfgen(int _n, const bam_pileup1_t *pl, int ref_base, bcf_callaux_t
         else
             bca->fwd_mqs[imq]++;
 
-        if ( bam_seqi(bam_get_seq(p->b),p->qpos) == ref_base )
+        if ( !is_diff )
         {
             bca->ref_pos[epos]++;
             bca->ref_bq[ibq]++;
             bca->ref_mq[imq]++;
             bca->ref_scl[sc_len]++;
+            if ( inm>=0 )
+            {
+                bca->ref_nm[inm]++;
+                if ( r->ref_nm ) r->ref_nm[inm]++;
+            }
         }
         else
         {
@@ -333,16 +527,21 @@ int bcf_call_glfgen(int _n, const bam_pileup1_t *pl, int ref_base, bcf_callaux_t
             bca->alt_bq[ibq]++;
             bca->alt_mq[imq]++;
             bca->alt_scl[sc_len]++;
+            if ( inm>=0 )
+            {
+                bca->alt_nm[inm]++;
+                if ( r->alt_nm ) r->alt_nm[inm]++;
+            }
         }
     }
 
     // Compensate for AD not being counted on low quality REF indel matches.
     if ( r->ADF && bca->ambig_reads==B2B_INC_AD0 )
     {
-        for (i=0; i<4; i++) // verify: are the counters ever non-zero for i!=0?
+        for (i=0; i<4; i++)
         {
-            r->ADR[i] += ADR_ref_missed[i];
-            r->ADF[i] += ADF_ref_missed[i];
+            r->ADR[0] += ADR_ref_missed[i];
+            r->ADF[0] += ADF_ref_missed[i];
         }
     }
     else if ( r->ADF && bca->ambig_reads==B2B_INC_AD )
@@ -359,9 +558,19 @@ int bcf_call_glfgen(int _n, const bam_pileup1_t *pl, int ref_base, bcf_callaux_t
             for (i=0; i<4; i++) r->ADF[i] += lroundf((float)dp_ambig * r->ADF[i]/dp);
     }
 
+    // Else consider downgrading bca->bases[] scores by AD vs AD_ref_missed
+    // ratios.  This is detrimental on Illumina, but beneficial on PacBio CCS.
+    // It's possibly related to the homopolyer error likelihoods or overall
+    // Indel accuracy.  Maybe tie this in to the -h option?
+
     r->ori_depth = ori_depth;
     // glfgen
     errmod_cal(bca->e, n, 5, bca->bases, r->p); // calculate PL of each genotype
+
+    // TODO: account for the number of unassigned reads.  If depth is 50,
+    // but AD is 5,7 then it may look like a variant but it probably
+    // should be low quality.
+
     return n;
 }
 
@@ -504,17 +713,18 @@ double mann_whitney_1947_cdf(int n, int m, int U)
 double calc_mwu_bias_cdf(int *a, int *b, int n)
 {
     int na = 0, nb = 0, i;
-    double U = 0, ties = 0;
+    double U = 0;
+    //double ties = 0;
     for (i=0; i<n; i++)
     {
         na += a[i];
         U  += a[i] * (nb + b[i]*0.5);
         nb += b[i];
-        if ( a[i] && b[i] )
-        {
-            double tie = a[i] + b[i];
-            ties += (tie*tie-1)*tie;
-        }
+        // if ( a[i] && b[i] )
+        // {
+        //     double tie = a[i] + b[i];
+        //     ties += (tie*tie-1)*tie;
+        // }
     }
     if ( !na || !nb ) return HUGE_VAL;
 
@@ -548,7 +758,8 @@ double calc_mwu_bias_cdf(int *a, int *b, int n)
 double calc_mwu_bias(int *a, int *b, int n, int left)
 {
     int na = 0, nb = 0, i;
-    double U = 0, ties = 0;
+    double U = 0;
+    // double ties = 0;
     for (i=0; i<n; i++)
     {
         if (!a[i]) {
@@ -561,8 +772,8 @@ double calc_mwu_bias(int *a, int *b, int n, int left)
             na += a[i];
             U  += a[i] * (nb + b[i]*0.5);
             nb += b[i];
-            double tie = a[i] + b[i];
-            ties += (tie*tie-1)*tie;
+            // double tie = a[i] + b[i];
+            // ties += (tie*tie-1)*tie;
         }
     }
     if ( !na || !nb ) return HUGE_VAL;
@@ -635,7 +846,7 @@ double calc_mwu_biasZ(int *a, int *b, int n, int left_only, int do_Z) {
         }
     }
 
-    if (na+nb <= 1)
+    if (!na || !nb)
         return HUGE_VAL;
 
     double U, m;
@@ -646,7 +857,7 @@ double calc_mwu_biasZ(int *a, int *b, int n, int left_only, int do_Z) {
     double var2 = (na*nb)/12.0 * ((na+nb+1) - t/(double)((na+nb)*(na+nb-1)));
     // var = na*nb*(na+nb+1)/12.0; // simpler; minus tie adjustment
     if (var2 <= 0)
-        return HUGE_VAL;
+        return do_Z ? 0 : 1;
 
     if (do_Z) {
         // S.D. normalised Z-score
@@ -798,6 +1009,7 @@ int bcf_call_combine(int n, const bcf_callret1_t *calls, bcf_callaux_t *bca, int
         call->n_alleles = j;
         if (call->n_alleles == 1) return -1; // no reliable supporting read. stop doing anything
     }
+    int has_alt = (call->n_alleles==2 && call->unseen!=-1) ? 0 : 1;
     /*
      * Set the phread likelihood array (call->PL) This array is 15 entries long
      * for each sample because that is size of an upper or lower triangle of a
@@ -900,6 +1112,7 @@ int bcf_call_combine(int n, const bcf_callret1_t *calls, bcf_callaux_t *bca, int
         }
 
 //      if (ref_base < 0) fprintf(bcftools_stderr, "%d,%d,%f,%d\n", call->n_alleles, x, sum_min, call->unseen);
+        // fprintf(bcftools_stderr,"sum_min=%f\n",sum_min);
         call->shift = (int)(sum_min + .499);
     }
     // combine annotations
@@ -914,59 +1127,50 @@ int bcf_call_combine(int n, const bcf_callret1_t *calls, bcf_callaux_t *bca, int
         for (j = 0; j < 16; ++j) call->anno[j] += calls[i].anno[j];
     }
 
-    calc_SegBias(calls, call);
+    // No need to calculate MWU tests when there is no ALT allele, this should speed up things slightly
+    if ( !has_alt ) return 0;
+
+    if ( bca->fmt_flag & B2B_INFO_FS )
+    {
+        double left,right,two;
+        call->strand_bias = kt_fisher_exact(call->anno[0], call->anno[1], call->anno[2], call->anno[3], &left, &right, &two);
+    }
+    if ( bca->fmt_flag & B2B_INFO_SGB ) calc_SegBias(calls, call);
 
     // calc_chisq_bias("XPOS", call->bcf_hdr->id[BCF_DT_CTG][call->tid].key, call->pos, bca->ref_pos, bca->alt_pos, bca->npos);
     // calc_chisq_bias("XMQ", call->bcf_hdr->id[BCF_DT_CTG][call->tid].key, call->pos, bca->ref_mq, bca->alt_mq, bca->nqual);
     // calc_chisq_bias("XBQ", call->bcf_hdr->id[BCF_DT_CTG][call->tid].key, call->pos, bca->ref_bq, bca->alt_bq, bca->nqual);
 
-    if (bca->fmt_flag & B2B_INFO_ZSCORE) {
-        // U z-normalised as +/- number of standard deviations from mean.
-        if (call->ori_ref < 0) {
-            if (bca->fmt_flag & B2B_INFO_RPB)
-                call->mwu_pos = calc_mwu_biasZ(bca->iref_pos, bca->ialt_pos,
-                                               bca->npos, 0, 1);
-            call->mwu_mq  = calc_mwu_biasZ(bca->iref_mq,  bca->ialt_mq,
-                                           bca->nqual,1,1);
-            if ( bca->fmt_flag & B2B_INFO_SCB )
-                call->mwu_sc  = calc_mwu_biasZ(bca->iref_scl, bca->ialt_scl,
-                                               100, 0,1);
-        } else {
-            if (bca->fmt_flag & B2B_INFO_RPB)
-                call->mwu_pos = calc_mwu_biasZ(bca->ref_pos, bca->alt_pos,
-                                               bca->npos, 0, 1);
-            call->mwu_mq  = calc_mwu_biasZ(bca->ref_mq,  bca->alt_mq,
-                                           bca->nqual,1,1);
-            call->mwu_bq  = calc_mwu_biasZ(bca->ref_bq,  bca->alt_bq,
-                                           bca->nqual,0,1);
-            call->mwu_mqs = calc_mwu_biasZ(bca->fwd_mqs, bca->rev_mqs,
-                                           bca->nqual,0,1);
-            if ( bca->fmt_flag & B2B_INFO_SCB )
-                call->mwu_sc  = calc_mwu_biasZ(bca->ref_scl, bca->alt_scl,
-                                               100, 0,1);
-        }
+    // U z-normalised as +/- number of standard deviations from mean.
+    if (call->ori_ref < 0) {    // indel
+        if ( bca->fmt_flag & B2B_INFO_RPBZ )
+            call->mwu_pos = calc_mwu_biasZ(bca->iref_pos, bca->ialt_pos, bca->npos, 0, 1);
+        if ( bca->fmt_flag & B2B_INFO_MQBZ )
+            call->mwu_mq  = calc_mwu_biasZ(bca->iref_mq,  bca->ialt_mq, bca->nqual,1,1);
+        if ( bca->fmt_flag & B2B_INFO_SCBZ )
+            call->mwu_sc  = calc_mwu_biasZ(bca->iref_scl, bca->ialt_scl, 100, 0,1);
     } else {
-        // Old method; U as probability between 0 and 1
-        if ( bca->fmt_flag & B2B_INFO_RPB )
-            call->mwu_pos = calc_mwu_biasZ(bca->ref_pos, bca->alt_pos,
-                                           bca->npos, 0, 0);
-        call->mwu_mq  = calc_mwu_biasZ(bca->ref_mq,  bca->alt_mq,
-                                       bca->nqual, 1, 0);
-        call->mwu_bq  = calc_mwu_biasZ(bca->ref_bq,  bca->alt_bq,
-                                       bca->nqual, 0, 0);
-        call->mwu_mqs = calc_mwu_biasZ(bca->fwd_mqs, bca->rev_mqs,
-                                       bca->nqual, 0, 0);
+        if ( bca->fmt_flag & B2B_INFO_RPBZ )
+            call->mwu_pos = calc_mwu_biasZ(bca->ref_pos, bca->alt_pos, bca->npos, 0, 1);
+        if ( bca->fmt_flag & B2B_INFO_MQBZ )
+            call->mwu_mq  = calc_mwu_biasZ(bca->ref_mq,  bca->alt_mq, bca->nqual,1,1);
+        if ( bca->fmt_flag & B2B_INFO_BQBZ )
+            call->mwu_bq  = calc_mwu_biasZ(bca->ref_bq,  bca->alt_bq, bca->nqual,0,1);
+        if ( bca->fmt_flag & B2B_INFO_MQSBZ )
+            call->mwu_mqs = calc_mwu_biasZ(bca->fwd_mqs, bca->rev_mqs, bca->nqual,0,1);
+        if ( bca->fmt_flag & B2B_INFO_SCBZ )
+            call->mwu_sc  = calc_mwu_biasZ(bca->ref_scl, bca->alt_scl, 100, 0,1);
+    }
+    if ( bca->fmt_flag & B2B_INFO_NMBZ )
+        call->mwu_nm[0] = calc_mwu_biasZ(bca->ref_nm, bca->alt_nm, B2B_N_NM,0,1);
+    if ( bca->fmt_flag & B2B_FMT_NMBZ )
+    {
+        for (i=0; i<n; i++)
+        {
+            float val = calc_mwu_biasZ(calls[i].ref_nm, calls[i].alt_nm, B2B_N_NM,0,1);
+            call->mwu_nm[i+1] = val!=HUGE_VAL ? val : 0;
+        }
     }
-
-#if CDF_MWU_TESTS
-    // CDF version of MWU tests is not calculated by default
-    if ( bca->fmt_flag & B2B_INFO_RPB )
-        call->mwu_pos_cdf = calc_mwu_bias_cdf(bca->ref_pos, bca->alt_pos, bca->npos);
-    call->mwu_mq_cdf  = calc_mwu_bias_cdf(bca->ref_mq,  bca->alt_mq,  bca->nqual);
-    call->mwu_bq_cdf  = calc_mwu_bias_cdf(bca->ref_bq,  bca->alt_bq,  bca->nqual);
-    call->mwu_mqs_cdf = calc_mwu_bias_cdf(bca->fwd_mqs, bca->rev_mqs, bca->nqual);
-#endif
-
     if ( bca->fmt_flag & B2B_INFO_VDB )
         call->vdb = calc_vdb(bca->alt_pos, bca->npos);
 
@@ -976,7 +1180,7 @@ int bcf_call_combine(int n, const bcf_callret1_t *calls, bcf_callaux_t *bca, int
 int bcf_call2bcf(bcf_call_t *bc, bcf1_t *rec, bcf_callret1_t *bcr, int fmt_flag, const bcf_callaux_t *bca, const char *ref)
 {
     extern double kt_fisher_exact(int n11, int n12, int n21, int n22, double *_left, double *_right, double *two);
-    int i, j, nals = 1;
+    int i, j, nals = 1, has_alt = 0;
 
     bcf_hdr_t *hdr = bc->bcf_hdr;
     rec->rid  = bc->tid;
@@ -1006,6 +1210,7 @@ int bcf_call2bcf(bcf_call_t *bc, bcf1_t *rec, bcf_callret1_t *bcr, int fmt_flag,
                 for (j = 0; j < bca->indelreg; ++j) kputc(ref[bc->pos+1+j], &bc->tmp);
             }
             nals++;
+            has_alt = 1;
         }
     }
     else    // SNP
@@ -1016,7 +1221,11 @@ int bcf_call2bcf(bcf_call_t *bc, bcf1_t *rec, bcf_callret1_t *bcr, int fmt_flag,
             if (bc->a[i] < 0) break;
             kputc(',', &bc->tmp);
             if ( bc->unseen==i ) kputs("<*>", &bc->tmp);
-            else kputc("ACGT"[bc->a[i]], &bc->tmp);
+            else
+            {
+                kputc("ACGT"[bc->a[i]], &bc->tmp);
+                has_alt = 1;
+            }
             nals++;
         }
     }
@@ -1025,11 +1234,33 @@ int bcf_call2bcf(bcf_call_t *bc, bcf1_t *rec, bcf_callret1_t *bcr, int fmt_flag,
     bc->tmp.l = 0;
 
     // INFO
-    if (bc->ori_ref < 0)
+    if ( bc->ori_ref < 0 )
     {
         bcf_update_info_flag(hdr, rec, "INDEL", NULL, 1);
-        bcf_update_info_int32(hdr, rec, "IDV", &bca->max_support, 1);
-        bcf_update_info_float(hdr, rec, "IMF", &bca->max_frac, 1);
+        uint32_t idv = bca->max_support;
+        if ( fmt_flag&B2B_INFO_IMF) {
+            float max_frac;
+            // Recompute IDV and IMF based on alignment results for more
+            // accurate counts, but only when in new "--indels-cns" mode.
+            if (bc->ADF && bc->ADR && bca->edlib) {
+                int max_ad = 0;
+                for (int k = 1; k < rec->n_allele; k++) {
+                    if (max_ad < bc->ADF[k] + bc->ADR[k])
+                        max_ad = bc->ADF[k] + bc->ADR[k];
+                }
+                max_frac = (double)(max_ad) / bc->ori_depth;
+                idv = max_ad;
+            } else {
+                max_frac = bca->max_frac;
+            }
+            // Copied here to maintain order for consistency of "make check"
+            if ( fmt_flag&B2B_INFO_IDV )
+                bcf_update_info_int32(hdr, rec, "IDV", &idv, 1);
+            bcf_update_info_float(hdr, rec, "IMF", &max_frac, 1);
+        } else {
+            if ( fmt_flag&B2B_INFO_IDV )
+                bcf_update_info_int32(hdr, rec, "IDV", &idv, 1);
+        }
     }
     bcf_update_info_int32(hdr, rec, "DP", &bc->ori_depth, 1);
     if ( fmt_flag&B2B_INFO_ADF )
@@ -1052,42 +1283,39 @@ int bcf_call2bcf(bcf_call_t *bc, bcf1_t *rec, bcf_callret1_t *bcr, int fmt_flag,
     bcf_update_info_float(hdr, rec, "I16", tmpf, 16);
     bcf_update_info_float(hdr, rec, "QS", bc->qsum, nals);
 
-    if ( bc->vdb != HUGE_VAL )      bcf_update_info_float(hdr, rec, "VDB", &bc->vdb, 1);
-    if ( bc->seg_bias != HUGE_VAL ) bcf_update_info_float(hdr, rec, "SGB", &bc->seg_bias, 1);
+    if ( has_alt )
+    {
+        if ( fmt_flag&B2B_INFO_MIN_PL_SUM )
+            bcf_update_info_int32(hdr, rec, "MIN_PL_SUM", &bc->shift, 1);
+        if ( fmt_flag&B2B_INFO_VDB && bc->vdb != HUGE_VAL )
+            bcf_update_info_float(hdr, rec, "VDB", &bc->vdb, 1);
+        if ( fmt_flag&B2B_INFO_SGB && bc->seg_bias != HUGE_VAL )
+            bcf_update_info_float(hdr, rec, "SGB", &bc->seg_bias, 1);
+        if ( fmt_flag&B2B_INFO_NM && (bca->nnm[0] || bca->nnm[1]) )
+        {
+            for (i=0; i<2; i++) bc->nm[i] = bca->nnm[i] ? bca->nm[i]/bca->nnm[i] : 0;
+            bcf_update_info_float(hdr, rec, "NM", bc->nm, 2);
+        }
 
-    if (bca->fmt_flag & B2B_INFO_ZSCORE) {
-        if ( bc->mwu_pos != HUGE_VAL )
+        if ( fmt_flag&B2B_INFO_RPBZ && bc->mwu_pos != HUGE_VAL )
             bcf_update_info_float(hdr, rec, "RPBZ", &bc->mwu_pos, 1);
-        if ( bc->mwu_mq != HUGE_VAL )
+        if ( fmt_flag&B2B_INFO_MQBZ && bc->mwu_mq != HUGE_VAL )
             bcf_update_info_float(hdr, rec, "MQBZ", &bc->mwu_mq, 1);
-        if ( bc->mwu_mqs != HUGE_VAL )
+        if ( fmt_flag&B2B_INFO_MQSBZ && bc->mwu_mqs != HUGE_VAL )
             bcf_update_info_float(hdr, rec, "MQSBZ", &bc->mwu_mqs, 1);
-        if ( bc->mwu_bq != HUGE_VAL )
+        if ( fmt_flag&B2B_INFO_BQBZ && bc->mwu_bq != HUGE_VAL )
             bcf_update_info_float(hdr, rec, "BQBZ", &bc->mwu_bq, 1);
-        if ( bc->mwu_sc != HUGE_VAL )
+        if ( fmt_flag&B2B_INFO_NMBZ && bc->mwu_nm[0] != HUGE_VAL )
+            bcf_update_info_float(hdr, rec, "NMBZ", bc->mwu_nm, 1);
+        if ( fmt_flag&B2B_INFO_SCBZ && bc->mwu_sc != HUGE_VAL )
             bcf_update_info_float(hdr, rec, "SCBZ", &bc->mwu_sc, 1);
-    } else {
-        if ( bc->mwu_pos != HUGE_VAL )
-            bcf_update_info_float(hdr, rec, "RPB", &bc->mwu_pos, 1);
-        if ( bc->mwu_mq != HUGE_VAL )
-            bcf_update_info_float(hdr, rec, "MQB", &bc->mwu_mq, 1);
-        if ( bc->mwu_mqs != HUGE_VAL )
-             bcf_update_info_float(hdr, rec, "MQSB", &bc->mwu_mqs, 1);
-        if ( bc->mwu_bq != HUGE_VAL )
-            bcf_update_info_float(hdr, rec, "BQB", &bc->mwu_bq, 1);
+        if ( fmt_flag&B2B_INFO_FS && bc->strand_bias != HUGE_VAL )
+            bcf_update_info_float(hdr, rec, "FS", &bc->strand_bias, 1);
     }
 
-    if ( bc->strand_bias != HUGE_VAL )
-        bcf_update_info_float(hdr, rec, "FS", &bc->strand_bias, 1);
-
-#if CDF_MWU_TESTS
-    if ( bc->mwu_pos_cdf != HUGE_VAL )  bcf_update_info_float(hdr, rec, "RPB2", &bc->mwu_pos_cdf, 1);
-    if ( bc->mwu_mq_cdf != HUGE_VAL )   bcf_update_info_float(hdr, rec, "MQB2", &bc->mwu_mq_cdf, 1);
-    if ( bc->mwu_mqs_cdf != HUGE_VAL )  bcf_update_info_float(hdr, rec, "MQSB2", &bc->mwu_mqs_cdf, 1);
-    if ( bc->mwu_bq_cdf != HUGE_VAL )   bcf_update_info_float(hdr, rec, "BQB2", &bc->mwu_bq_cdf, 1);
-#endif
     tmpf[0] = bc->ori_depth ? (float)bc->mq0/bc->ori_depth : 0;
-    bcf_update_info_float(hdr, rec, "MQ0F", tmpf, 1);
+    if ( fmt_flag&B2B_INFO_MQ0F )
+        bcf_update_info_float(hdr, rec, "MQ0F", tmpf, 1);
 
     // FORMAT
     rec->n_sample = bc->n;
@@ -1144,5 +1372,11 @@ int bcf_call2bcf(bcf_call_t *bc, bcf1_t *rec, bcf_callret1_t *bcr, int fmt_flag,
     if ( fmt_flag&B2B_FMT_QS )
         bcf_update_format_int32(hdr, rec, "QS", bc->QS, rec->n_sample*rec->n_allele);
 
+    if ( has_alt )
+    {
+        if ( fmt_flag&B2B_FMT_NMBZ )
+            bcf_update_format_float(hdr, rec, "NMBZ", bc->mwu_nm+1, rec->n_sample);
+    }
+
     return 0;
 }
