@@ -2,7 +2,7 @@
 
 /*  mpileup.c -- mpileup subcommand. Previously bam_plcmd.c from samtools
 
-    Copyright (C) 2008-2022 Genome Research Ltd.
+    Copyright (C) 2008-2024 Genome Research Ltd.
     Portions copyright (C) 2009-2012 Broad Institute.
 
     Author: Heng Li <lh3@sanger.ac.uk>
@@ -74,8 +74,11 @@ typedef struct {
     uint32_t fmt_flag;
     int rflag_skip_any_unset, rflag_skip_all_unset, rflag_skip_any_set, rflag_skip_all_set, output_type;
     int openQ, extQ, tandemQ, min_support, indel_win_size; // for indels
+    int seqQ_offset;
     double min_frac; // for indels
-    double indel_bias;
+    double indel_bias, poly_mqual;
+    double del_bias; // compensate for diff deletion vs insertion error rates
+    double vs_ref;
     char *reg_fname, *pl_list, *fai_fname, *output_fname;
     int reg_is_file, record_cmd_line, n_threads, clevel;
     faidx_t *fai;
@@ -101,8 +104,11 @@ typedef struct {
     htsFile *bcf_fp;
     bcf_hdr_t *bcf_hdr;
     int indels_v20;
+    int edlib;
     int argc;
     char **argv;
+    int write_index;
+    char *index_fn;
 } mplp_conf_t;
 
 typedef struct {
@@ -491,37 +497,43 @@ static void mplp_realn(int n, int *n_plp, const bam_pileup1_t **plp,
             if ((flag & MPLP_REALN_PARTIAL) && nt > 15 && ncig > 1) {
                 // Left & right cigar op match.
                 int lr = b->core.l_qseq > 500;
-                int lm = 0, rm = 0, k;
+                int lm = 0, rm = 0, k, nm = 0;
                 for (k = 0; k < ncig; k++) {
                     int cop = bam_cigar_op(cig[k]);
                     if (lr && (cop == BAM_CHARD_CLIP || cop == BAM_CSOFT_CLIP))
                         continue;
 
                     if (cop == BAM_CMATCH || cop == BAM_CDIFF ||
-                        cop == BAM_CEQUAL)
+                        cop == BAM_CEQUAL) {
                         lm += bam_cigar_oplen(cig[k]);
-                    else
+                        nm++;
+                    } else {
                         break;
+                    }
                 }
 
-                for (k = ncig-1; k >= 0; k--) {
-                    int cop = bam_cigar_op(cig[k]);
-                    if (lr && (cop == BAM_CHARD_CLIP || cop == BAM_CSOFT_CLIP))
+                // if everything is a match (or sequence (mis)match) then move on
+                // because we don't have an indel in the middle
+                if (nm != ncig) {
+                    for (k = ncig-1; k >= 0; k--) {
+                        int cop = bam_cigar_op(cig[k]);
+                        if (lr && (cop == BAM_CHARD_CLIP || cop == BAM_CSOFT_CLIP))
+                            continue;
+
+                        if (cop == BAM_CMATCH || cop == BAM_CDIFF ||
+                            cop == BAM_CEQUAL)
+                            rm += bam_cigar_oplen(cig[k]);
+                        else
+                            break;
+                    }
+
+                    if (lm >= REALN_DIST*4 && rm >= REALN_DIST*4)
                         continue;
 
-                    if (cop == BAM_CMATCH || cop == BAM_CDIFF ||
-                        cop == BAM_CEQUAL)
-                        rm += bam_cigar_oplen(cig[k]);
-                    else
-                        break;
+                    if (lm >= REALN_DIST && rm >= REALN_DIST &&
+                        has_clip < (0.15+0.05*(nt>20))*nt)
+                        continue;
                 }
-
-                if (lm >= REALN_DIST*4 && rm >= REALN_DIST*4)
-                    continue;
-
-                if (lm >= REALN_DIST && rm >= REALN_DIST &&
-                    has_clip < (0.15+0.05*(nt>20))*nt)
-                    continue;
             }
 
             if (b->core.l_qseq > 500) {
@@ -578,12 +590,14 @@ static int mpileup_reg(mplp_conf_t *conf, uint32_t beg, uint32_t end)
 
         // call indels; todo: subsampling with total_depth>max_indel_depth instead of ignoring?
         // check me: rghash in bcf_call_gap_prep() should have no effect, reads mplp_func already excludes them
-        if ( !(conf->flag&MPLP_NO_INDEL) && total_depth < conf->max_indel_depth )
+        if (!(conf->flag&MPLP_NO_INDEL) && total_depth < conf->max_indel_depth)
         {
             bcf_callaux_clean(conf->bca, &conf->bc);
             conf->bca->chr = tid>=0 ? hdr->target_name[tid] : NULL;
             int iret;
-            if ( conf->indels_v20 )
+            if (conf->edlib)
+                iret = bcf_edlib_gap_prep(conf->gplp->n, conf->gplp->n_plp, conf->gplp->plp, pos, conf->bca, ref, ref_len);
+            else if ( conf->indels_v20 )
                 iret = bcf_iaux_gap_prep(conf->gplp->n, conf->gplp->n_plp, conf->gplp->plp, pos, conf->bca, ref);
             else
                 iret = bcf_call_gap_prep(conf->gplp->n, conf->gplp->n_plp, conf->gplp->plp, pos, conf->bca, ref);
@@ -600,7 +614,7 @@ static int mpileup_reg(mplp_conf_t *conf, uint32_t beg, uint32_t end)
             }
         }
     }
-    return 0;
+    return ret;
 }
 
 static int mpileup(mplp_conf_t *conf)
@@ -640,8 +654,12 @@ static int mpileup(mplp_conf_t *conf)
             }
         }
         nregs = regidx_nregs(conf->reg);
-        conf->reg_itr = regitr_init(conf->reg);
-        regitr_loop(conf->reg_itr);   // region iterator now positioned at the first region
+        if ( nregs )
+        {
+            // the regions list can be empty, see #2250
+            conf->reg_itr = regitr_init(conf->reg);
+            regitr_loop(conf->reg_itr);   // region iterator now positioned at the first region
+        }
     }
 
     // read the header of each file in the list and initialize data
@@ -687,7 +705,7 @@ static int mpileup(mplp_conf_t *conf)
             i--;
             continue;
         }
-        if (conf->reg) {
+        if (conf->reg && nregs) {
             hts_idx_t *idx = sam_index_load(conf->mplp_data[i]->fp, conf->files[i]);
             if (idx == NULL) {
                 fprintf(bcftools_stderr, "[%s] fail to load index for %s\n", __func__, conf->files[i]);
@@ -851,12 +869,16 @@ static int mpileup(mplp_conf_t *conf)
     for (i=0; i<nsmpl; i++)
         bcf_hdr_add_sample(conf->bcf_hdr, smpl[i]);
     if ( bcf_hdr_write(conf->bcf_fp, conf->bcf_hdr)!=0 ) error("[%s] Error: failed to write the header to %s\n",__func__,conf->output_fname?conf->output_fname:"standard output");
+    if ( init_index2(conf->bcf_fp,conf->bcf_hdr,conf->output_fname,
+                     &conf->index_fn, conf->write_index) < 0 )
+        error("Error: failed to initialise index for %s\n",conf->output_fname);
 
     conf->bca = bcf_call_init(-1., conf->min_baseQ, conf->max_baseQ,
                               conf->delta_baseQ);
     conf->bcr = (bcf_callret1_t*) calloc(nsmpl, sizeof(bcf_callret1_t));
     conf->bca->openQ = conf->openQ, conf->bca->extQ = conf->extQ, conf->bca->tandemQ = conf->tandemQ;
     conf->bca->indel_bias = conf->indel_bias;
+    conf->bca->del_bias = conf->del_bias;
     conf->bca->min_frac = conf->min_frac;
     conf->bca->min_support = conf->min_support;
     conf->bca->per_sample_flt = conf->flag & MPLP_PER_SAMPLE;
@@ -864,6 +886,10 @@ static int mpileup(mplp_conf_t *conf)
     conf->bca->ambig_reads = conf->ambig_reads;
     conf->bca->indel_win_size = conf->indel_win_size;
     conf->bca->indels_v20 = conf->indels_v20;
+    conf->bca->edlib = conf->edlib;
+    conf->bca->seqQ_offset = conf->seqQ_offset;
+    conf->bca->poly_mqual = conf->poly_mqual;
+    conf->bca->vs_ref = conf->vs_ref;
 
     conf->bc.bcf_hdr = conf->bcf_hdr;
     conf->bc.n  = nsmpl;
@@ -918,6 +944,7 @@ static int mpileup(mplp_conf_t *conf)
 
 
     // Run mpileup for multiple regions
+    int ret = 0;
     if ( nregs )
     {
         int ireg = 0;
@@ -946,12 +973,18 @@ static int mpileup(mplp_conf_t *conf)
                     bam_mplp_reset(conf->iter);
                 }
             }
-            mpileup_reg(conf,conf->reg_itr->beg,conf->reg_itr->end);
+            ret = mpileup_reg(conf,conf->reg_itr->beg,conf->reg_itr->end);
+            if ( ret<0 ) break;
         }
         while ( regitr_loop(conf->reg_itr) );
     }
-    else
-        mpileup_reg(conf,0,UINT32_MAX);
+    else if ( !conf->reg )
+        ret = mpileup_reg(conf,0,UINT32_MAX);
+    if ( ret<0 )
+    {
+        fprintf(bcftools_stderr, "[%s] failed to read from input file\n", __func__);
+        bcftools_exit(EXIT_FAILURE);
+    }
 
     flush_bcf_records(conf, conf->bcf_fp, conf->bcf_hdr, NULL);
 
@@ -960,6 +993,15 @@ static int mpileup(mplp_conf_t *conf)
     bcf_destroy1(conf->bcf_rec);
     if (conf->bcf_fp)
     {
+        if ( conf->write_index )
+        {
+            if ( bcf_idx_save(conf->bcf_fp)<0 )
+            {
+                if ( hts_close(conf->bcf_fp)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,conf->output_fname);
+                error("Error: cannot write to index %s\n",conf->index_fn);
+            }
+            free(conf->index_fn);
+        }
         if ( hts_close(conf->bcf_fp)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,conf->output_fname);
         bcf_hdr_destroy(conf->bcf_hdr);
         bcf_call_destroy(conf->bca);
@@ -1134,7 +1176,7 @@ static void list_annotations(FILE *fp)
         "\n"
         "FORMAT annotation tags available (\"FORMAT/\" prefix is optional):\n"
         "\n"
-        "  FORMAT/AD   .. Allelic depth (Number=R,Type=Integer)\n"
+        "* FORMAT/AD   .. Allelic depth (Number=R,Type=Integer)\n"
         "  FORMAT/ADF  .. Allelic depths on the forward strand (Number=R,Type=Integer)\n"
         "  FORMAT/ADR  .. Allelic depths on the reverse strand (Number=R,Type=Integer)\n"
         "  FORMAT/DP   .. Number of high-quality bases (Number=1,Type=Integer)\n"
@@ -1184,7 +1226,7 @@ static void print_usage(FILE *fp, const mplp_conf_t *mplp)
         "\n"
         "Input options:\n"
         "  -6, --illumina1.3+      Quality is in the Illumina-1.3+ encoding\n"
-        "  -A, --count-orphans     Do not discard anomalous read pairs\n"
+        "  -A, --count-orphans     Include anomalous read pairs, with flag PAIRED but not PROPER_PAIR set\n"
         "  -b, --bam-list FILE     List of input BAM filenames, one per line\n"
         "  -B, --no-BAQ            Disable BAQ (per-Base Alignment Quality)\n"
         "  -C, --adjust-MQ INT     Adjust mapping quality [0]\n"
@@ -1229,9 +1271,10 @@ static void print_usage(FILE *fp, const mplp_conf_t *mplp)
         "  -O, --output-type TYPE  'b' compressed BCF; 'u' uncompressed BCF;\n"
         "                          'z' compressed VCF; 'v' uncompressed VCF; 0-9 compression level [v]\n"
         "      --threads INT       Use multithreading with INT worker threads [0]\n"
+        "  -W, --write-index[=FMT] Automatically index the output files [off]\n"
         "\n"
         "SNP/INDEL genotype likelihoods options:\n"
-        "  -X, --config STR        Specify platform specific profiles (see below)\n"
+        "  -X, --config STR        Specify platform profile (use \"-X list\" for details)\n"
         "  -e, --ext-prob INT      Phred-scaled gap extension seq error probability [%d]\n", mplp->extQ);
     fprintf(fp,
         "  -F, --gap-frac FLOAT    Minimum fraction of gapped reads [%g]\n", mplp->min_frac);
@@ -1252,24 +1295,26 @@ static void print_usage(FILE *fp, const mplp_conf_t *mplp)
         "  --ar, --ambig-reads STR   What to do with ambiguous indel reads: drop,incAD,incAD0 [drop]\n");
     fprintf(fp,
         "      --indel-bias FLOAT  Raise to favour recall over precision [%.2f]\n", mplp->indel_bias);
+    fprintf(fp,
+        "      --del-bias FLOAT    Relative likelihood of insertion to deletion [%.2f]\n", mplp->del_bias);
+    fprintf(fp,
+        "      --score-vs-ref FLOAT\n"
+        "                          Ratio of score vs ref (1) or 2nd-best allele (0) [%.2f]\n", mplp->vs_ref);
     fprintf(fp,
         "      --indel-size INT    Approximate maximum indel size considered [%d]\n", mplp->indel_win_size);
     fprintf(fp,
-        "      --indels-2.0        New EXPERIMENTAL indel calling model (diploid reference consensus)\n");
+        "      --indels-2.0        New EXPERIMENTAL indel calling model (diploid reference consensus)\n"
+        "      --indels-cns        New EXPERIMENTAL indel calling model with edlib\n"
+        "      --seqq-offset       Indel-cns tuning for indel seq-qual scores [120]\n"
+        "      --no-indels-cns     Disable CNS mode, to use after a -X profile\n"
+        "      --poly-mqual        (Edlib mode) Use minimum quality within homopolymers\n");
     fprintf(fp,"\n");
     fprintf(fp,
-        "Configuration profiles activated with -X, --config:\n"
-        "    1.12:        -Q13 -h100 -m1 -F0.002\n"
-        "    illumina:    [ default values ]\n"
-        "    ont:         -B -Q5 --max-BQ 30 -I [also try eg |bcftools call -P0.01]\n"
-        "    pacbio-ccs:  -D -Q5 --max-BQ 50 -F0.1 -o25 -e1 --delta-BQ 10 -M99999\n"
-        "\n"
-        "Notes: Assuming diploid individuals.\n"
-        "\n"
-        "Example:\n"
-        "   # See also http://samtools.github.io/bcftools/howtos/variant-calling.html\n"
-        "   bcftools mpileup -Ou -f reference.fa alignments.bam | bcftools call -mv -Ob -o calls.bcf\n"
-        "\n");
+            "Notes: Assuming diploid individuals.\n\n"
+            "Example:\n"
+            "   # See also http://samtools.github.io/bcftools/howtos/variant-calling.html\n"
+            "   bcftools mpileup -Ou -f reference.fa alignments.bam | bcftools call -mv -Ob -o calls.bcf\n"
+            "\n");
 
     free(tmp_skip_all_set);
     free(tmp_skip_any_unset);
@@ -1277,9 +1322,41 @@ static void print_usage(FILE *fp, const mplp_conf_t *mplp)
     free(tmp_skip_any_set);
 }
 
+static void print_profiles(void) {
+    fprintf(bcftools_stdout, 
+"Configuration profiles activated with -X, --config:\n\n"
+"1.12\n"
+"    -Q13 -h100 -m1 -F0.002\n\n"
+"bgi, bgi-1.20\n"
+"    --indels-cns -B --indel-size 80 -F0.1 --indel-bias 0.9 --seqq-offset 120\n\n"
+"illumina-1.18\n"
+"    --indel-size 110\n\n"
+"illumina\n"
+"illumina-1.20\n"
+"    --indels-cns --indel-size 110\n\n"
+"ont\n"
+"    -B -Q5 --max-BQ 30 -I\n\n"
+"ont-sup, ont-sup-1.20\n"
+"    --indels-cns -B -Q1 --max-BQ 35 -F0.2 -o15 -e1 -h110 --delta-BQ 99\\\n"
+"    --del-bias 0.4 --indel-bias 0.7 --poly-mqual --seqq-offset 130\\\n"
+"    --indel-size 80\n\n"
+"pacbio-ccs-1.18\n"
+"    -D -Q5 --max-BQ 50 -F0.1 -o25 -e1 --delta-BQ 10 \\\n"
+"    -M99999 --indel-size 110\n\n"
+"pacbio-ccs, pacbio-ccs-1.20\n"
+"    --indels-cns -B -Q5 --max-BQ 50 -F0.1 -o25 -e1 -h300 --delta-BQ 10 \\\n"
+"    --del-bias 0.4 --poly-mqual --indel-bias 0.9 --seqq-offset 118\\\n"
+"    --indel-size 80 --score-vs-ref 0.7\n\n"
+"ultima, ultima-1.20\n"
+"    --indels-cns -B -Q1 --max-BQ 30 -F0.15 -o20 -e10 -h250 --delta-BQ 10 \\\n"
+"    --del-bias 0.3 --indel-bias 0.7 --poly-mqual --seqq-offset 140 \\\n"
+"    --indel-size 80 --score-vs-ref 0.3\n\n"
+"\n");
+}
+
 int main_mpileup(int argc, char *argv[])
 {
-    int c;
+    int c, i, ret = 1;
     const char *file_list = NULL;
     char **fn = NULL;
     int nfiles = 0, use_orphan = 0, noref = 0;
@@ -1292,6 +1369,7 @@ int main_mpileup(int argc, char *argv[])
     mplp.max_depth = 250; mplp.max_indel_depth = 250;
     mplp.openQ = 40; mplp.extQ = 20; mplp.tandemQ = 500;
     mplp.min_frac = 0.05; mplp.indel_bias = 1.0; mplp.min_support = 2;
+    mplp.vs_ref = 0;
     mplp.flag = MPLP_NO_ORPHAN | MPLP_REALN | MPLP_REALN_PARTIAL
               | MPLP_SMART_OVERLAPS;
     mplp.argc = argc; mplp.argv = argv;
@@ -1302,11 +1380,14 @@ int main_mpileup(int argc, char *argv[])
     mplp.n_threads = 0;
     mplp.bsmpl = bam_smpl_init();
     // the default to be changed in future, see also parse_format_flag()
-    mplp.fmt_flag = B2B_INFO_BQBZ|B2B_INFO_IDV|B2B_INFO_IMF|B2B_INFO_MQ0F|B2B_INFO_MQBZ|B2B_INFO_MQSBZ|B2B_INFO_RPBZ|B2B_INFO_SCBZ|B2B_INFO_SGB|B2B_INFO_VDB;
+    mplp.fmt_flag = B2B_INFO_BQBZ|B2B_INFO_IDV|B2B_INFO_IMF|B2B_INFO_MQ0F|B2B_INFO_MQBZ|B2B_INFO_MQSBZ|B2B_INFO_RPBZ|B2B_INFO_SCBZ|B2B_INFO_SGB|B2B_INFO_VDB|B2B_FMT_AD;
     mplp.max_read_len = 500;
     mplp.ambig_reads = B2B_DROP;
     mplp.indel_win_size = 110;
+    mplp.poly_mqual = 0;
+    mplp.seqQ_offset = 120;
     mplp.clevel = -1;
+    mplp.del_bias = 0; // even insertion and deletion likelhoods.
     hts_srand48(0);
 
     static const struct option lopts[] =
@@ -1365,6 +1446,8 @@ int main_mpileup(int argc, char *argv[])
         {"indel-bias", required_argument, NULL, 10},
         {"indel-size", required_argument, NULL, 15},
         {"indels-2.0", no_argument, NULL, 20},
+        {"indels-cns", no_argument, NULL, 22},
+        {"no-indels-cns", no_argument, NULL, 25},
         {"tandem-qual", required_argument, NULL, 'h'},
         {"skip-indels", no_argument, NULL, 'I'},
         {"max-idepth", required_argument, NULL, 'L'},
@@ -1377,26 +1460,44 @@ int main_mpileup(int argc, char *argv[])
         {"seed", required_argument, NULL, 13},
         {"ambig-reads", required_argument, NULL, 14},
         {"ar", required_argument, NULL, 14},
+        {"write-index",optional_argument,NULL,'W'},
+        {"del-bias", required_argument, NULL, 23},
+        {"poly-mqual", no_argument, NULL, 24},
+        {"no-poly-mqual", no_argument, NULL, 26},
+        {"score-vs-ref",required_argument, NULL, 27},
+        {"seqq-offset", required_argument, NULL, 28},
         {NULL, 0, NULL, 0}
     };
-    while ((c = getopt_long(argc, argv, "Ag:f:r:R:q:Q:C:BDd:L:b:P:po:e:h:Im:F:EG:6O:xa:s:S:t:T:M:X:U",lopts,NULL)) >= 0) {
+    while ((c = getopt_long(argc, argv, "Ag:f:r:R:q:Q:C:BDd:L:b:P:po:e:h:Im:F:EG:6O:xa:s:S:t:T:M:X:UW::",lopts,NULL)) >= 0) {
         switch (c) {
         case 'x': mplp.flag &= ~MPLP_SMART_OVERLAPS; break;
         case  16 :
             mplp.rflag_skip_any_unset = bam_str2flag(optarg);
-            if ( mplp.rflag_skip_any_unset <0 ) { fprintf(bcftools_stderr,"Could not parse --nf %s\n", optarg); return 1; }
+            if ( mplp.rflag_skip_any_unset <0 ) {
+                fprintf(bcftools_stderr,"Could not parse --nf %s\n", optarg);
+                goto err;
+            }
             break;
         case  17 :
             mplp.rflag_skip_all_unset = bam_str2flag(optarg);
-            if ( mplp.rflag_skip_all_unset<0 ) { fprintf(bcftools_stderr,"Could not parse --if %s\n", optarg); return 1; }
+            if ( mplp.rflag_skip_all_unset<0 ) {
+                fprintf(bcftools_stderr,"Could not parse --if %s\n", optarg);
+                goto err;
+            }
             break;
         case  18 :
             mplp.rflag_skip_any_set = bam_str2flag(optarg);
-            if ( mplp.rflag_skip_any_set <0 ) { fprintf(bcftools_stderr,"Could not parse --ef %s\n", optarg); return 1; }
+            if ( mplp.rflag_skip_any_set <0 ) {
+                fprintf(bcftools_stderr,"Could not parse --ef %s\n", optarg);
+                goto err;
+            }
             break;
         case  19 :
             mplp.rflag_skip_all_set = bam_str2flag(optarg);
-            if ( mplp.rflag_skip_all_set <0 ) { fprintf(bcftools_stderr,"Could not parse --df %s\n", optarg); return 1; }
+            if ( mplp.rflag_skip_all_set <0 ) {
+                fprintf(bcftools_stderr,"Could not parse --df %s\n", optarg);
+                goto err;
+            }
             break;
         case  3 : mplp.output_fname = optarg; break;
         case  4 : mplp.openQ = atoi(optarg); break;
@@ -1407,7 +1508,8 @@ int main_mpileup(int argc, char *argv[])
             break;
         case 'f':
             mplp.fai = fai_load(optarg);
-            if (mplp.fai == NULL) return 1;
+            if (mplp.fai == NULL)
+                goto err;
             mplp.fai_fname = optarg;
             break;
         case  7 : noref = 1; break;
@@ -1434,7 +1536,10 @@ int main_mpileup(int argc, char *argv[])
                   if ( optarg[0]=='^' ) optarg++;
                   else mplp.bed_logic = 1;
                   mplp.bed = regidx_init(optarg,NULL,NULL,0,NULL);
-                  if (!mplp.bed) { fprintf(bcftools_stderr, "bcftools mpileup: Could not read file \"%s\"", optarg); return 1; }
+                  if (!mplp.bed) {
+                      fprintf(bcftools_stderr, "bcftools mpileup: Could not read file \"%s\"", optarg);
+                      goto err;
+                  }
                   break;
         case 'P': mplp.pl_list = strdup(optarg); break;
         case 'p': mplp.flag |= MPLP_PER_SAMPLE; break;
@@ -1487,18 +1592,39 @@ int main_mpileup(int argc, char *argv[])
             else
                 mplp.indel_bias = 1/atof(optarg);
             break;
+        case 27:
+            mplp.vs_ref = atof(optarg);
+            //if (mplp.vs_ref < 0) mplp.vs_ref = 0;
+            if (mplp.vs_ref > 1) mplp.vs_ref = 1;
+            break;
         case  15: {
                 char *tmp;
                 mplp.indel_win_size = strtol(optarg,&tmp,10);
                 if ( *tmp ) error("Could not parse argument: --indel-size %s\n", optarg);
-                if ( mplp.indel_win_size < 110 )
+                if ( mplp.indel_win_size < 20 )
                 {
-                    mplp.indel_win_size = 110;
+                    mplp.indel_win_size = 20;
                     fprintf(bcftools_stderr,"Warning: running with --indel-size %d, the requested value is too small\n",mplp.indel_win_size);
                 }
             }
             break;
-        case  20: mplp.indels_v20 = 1; break;
+        case  20: mplp.indels_v20 = 1; mplp.edlib = 0; break;
+        case 'W':
+            if (!(mplp.write_index = write_index_parse(optarg)))
+                error("Unsupported index format '%s'\n", optarg);
+            break;
+        case  22: mplp.edlib = 1; mplp.indels_v20 = 0; break;
+        case  25: mplp.edlib = 0; break;
+        case  28:
+            mplp.seqQ_offset = atoi(optarg);
+            if (mplp.seqQ_offset < 100)
+                mplp.seqQ_offset = 100;
+            if (mplp.seqQ_offset > 200)
+                mplp.seqQ_offset = 200;
+            break;
+        case  23: mplp.del_bias = atof(optarg); break;
+        case  24: mplp.poly_mqual = 1; break;
+        case  26: mplp.poly_mqual = 0; break;
         case 'A': use_orphan = 1; break;
         case 'F': mplp.min_frac = atof(optarg); break;
         case 'm': mplp.min_support = atoi(optarg); break;
@@ -1507,13 +1633,13 @@ int main_mpileup(int argc, char *argv[])
         case 'a':
             if (optarg[0]=='?') {
                 list_annotations(bcftools_stderr);
-                return 1;
+                goto err;
             }
             parse_format_flag(&mplp.fmt_flag,optarg);
         break;
         case 'M': mplp.max_read_len = atoi(optarg); break;
         case 'X':
-            if (strcasecmp(optarg, "pacbio-ccs") == 0) {
+            if (strcasecmp(optarg, "pacbio-ccs-1.18") == 0) {
                 mplp.min_frac = 0.1;
                 mplp.min_baseQ = 5;
                 mplp.max_baseQ = 50;
@@ -1522,13 +1648,70 @@ int main_mpileup(int argc, char *argv[])
                 mplp.extQ = 1;
                 mplp.flag |= MPLP_REALN_PARTIAL;
                 mplp.max_read_len = 99999;
+
+            } else if (strcasecmp(optarg, "pacbio-ccs") == 0 ||
+                strcasecmp(optarg, "pacbio-ccs-1.20") == 0) {
+                mplp.min_frac = 0.1;
+                mplp.min_baseQ = 5;
+                mplp.max_baseQ = 50;
+                mplp.delta_baseQ = 10;
+                mplp.tandemQ = 300;
+                mplp.openQ = 25;
+                mplp.extQ = 1;
+                mplp.flag &= ~MPLP_REALN;
+                mplp.del_bias = 0.4;
+                mplp.indel_bias = 1/.9;
+                mplp.seqQ_offset = 118;
+                mplp.poly_mqual = 1;
+                mplp.edlib = 1;
+                mplp.vs_ref = 0.7;
+                mplp.indel_win_size = 80;
+
             } else if (strcasecmp(optarg, "ont") == 0) {
-                fprintf(bcftools_stderr, "For ONT it may be beneficial to also run bcftools call with "
+                fprintf(bcftools_stderr, "With old ONT data may be beneficial to also run bcftools call with "
                         "a higher -P, eg -P0.01 or -P 0.1\n");
                 mplp.min_baseQ = 5;
                 mplp.max_baseQ = 30;
                 mplp.flag &= ~MPLP_REALN;
                 mplp.flag |= MPLP_NO_INDEL;
+
+            } else if (strcasecmp(optarg, "ont-sup") == 0 ||
+                       strcasecmp(optarg, "ont-sup-1.20") == 0) {
+                mplp.min_frac = 0.2;
+                mplp.min_baseQ = 1;
+                mplp.max_baseQ = 35;
+                mplp.delta_baseQ = 99;
+                mplp.openQ = 15;
+                mplp.extQ = 1;
+                mplp.flag &= ~MPLP_REALN;
+                mplp.max_read_len = 9999999;
+                mplp.del_bias = 0.4;
+                mplp.poly_mqual = 1;
+                mplp.edlib = 1;
+                // If we increase -h then we can increase bias denominator too
+                mplp.tandemQ = 110;
+                mplp.indel_bias = 1/0.7;
+                mplp.seqQ_offset = 130;
+                mplp.indel_win_size = 80;
+
+            } else if (strcasecmp(optarg, "ultima") == 0 ||
+                       strcasecmp(optarg, "ultima-1.20") == 0) {
+                mplp.min_frac = 0.15;
+                mplp.min_baseQ = 1;
+                mplp.max_baseQ = 30;
+                mplp.delta_baseQ = 10;
+                mplp.openQ = 20;
+                mplp.extQ = 10;
+                mplp.tandemQ = 250;
+                mplp.flag &= ~MPLP_REALN;
+                mplp.del_bias = 0.3;
+                mplp.poly_mqual = 1;
+                mplp.edlib = 1;
+                mplp.indel_bias = 1/0.7;
+                mplp.seqQ_offset = 140;
+                mplp.vs_ref = 0.3;
+                mplp.indel_win_size = 80;
+
             } else if (strcasecmp(optarg, "1.12") == 0) {
                 // 1.12 and earlier
                 mplp.min_frac = 0.002;
@@ -1537,13 +1720,38 @@ int main_mpileup(int argc, char *argv[])
                 mplp.tandemQ = 100;
                 mplp.flag &= ~MPLP_REALN_PARTIAL;
                 mplp.flag |= MPLP_REALN;
-            } else if (strcasecmp(optarg, "illumina") == 0) {
+
+            } else if (strcasecmp(optarg, "illumina-1.18") == 0) {
+                mplp.indel_win_size = 110;
                 mplp.flag |= MPLP_REALN_PARTIAL;
+
+            } else if (strcasecmp(optarg, "illumina") == 0 ||
+                       strcasecmp(optarg, "illumina-1.20") == 0) {
+                mplp.edlib = 1;
+                mplp.indel_win_size = 110;
+                mplp.flag |= MPLP_REALN_PARTIAL;
+                mplp.indel_bias = 1;
+                mplp.seqQ_offset = 125;
+                //mplp.indel_win_size = 80; TEST?
+
+            } else if (strcasecmp(optarg, "bgi") == 0 ||
+                       strcasecmp(optarg, "bgi-1.20") == 0) {
+                mplp.min_frac = 0.1;
+                mplp.edlib = 1;
+                mplp.indel_bias = 1;
+                mplp.seqQ_offset = 120;
+                mplp.flag |= MPLP_REALN_PARTIAL;
+                mplp.indel_win_size = 80;
+
+            } else if (strcasecmp(optarg, "list") == 0 ||
+                       strcasecmp(optarg, "help") == 0) {
+                print_profiles();
+                goto err;
             } else {
                 fprintf(bcftools_stderr, "Unknown configuration name '%s'\n"
-                        "Please choose from 1.12, illumina, pacbio-ccs or ont\n",
+                        "Please use '-X list' to show available choices.\n",
                         optarg);
-                return 1;
+                goto err;
             }
             break;
         case 13: hts_srand48(atoi(optarg)); break;
@@ -1555,7 +1763,7 @@ int main_mpileup(int argc, char *argv[])
             break;
         default:
             fprintf(bcftools_stderr,"Invalid option: '%c'\n", c);
-            return 1;
+            goto err;
         }
     }
 
@@ -1580,22 +1788,23 @@ int main_mpileup(int argc, char *argv[])
     if ( !(mplp.flag&MPLP_REALN) && mplp.flag&MPLP_REDO_BAQ )
     {
         fprintf(bcftools_stderr,"Error: The -B option cannot be combined with -E\n");
-        return 1;
+        goto err;
     }
     if (use_orphan) mplp.flag &= ~MPLP_NO_ORPHAN;
     if (argc == 1)
     {
         print_usage(bcftools_stderr, &mplp);
-        return 1;
+        goto err;
     }
     if (!mplp.fai && !noref) {
         fprintf(bcftools_stderr,"Error: mpileup requires the --fasta-ref option by default; use --no-reference to run without a fasta reference\n");
-        return 1;
+        goto err;
     }
-    int ret,i;
+
     if (file_list)
     {
-        if ( read_file_list(file_list,&nfiles,&fn) ) return 1;
+        if ( read_file_list(file_list,&nfiles,&fn) )
+            goto err;
         mplp.files  = fn;
         mplp.nfiles = nfiles;
     }
@@ -1614,6 +1823,8 @@ int main_mpileup(int argc, char *argv[])
     if (mplp.bed) regidx_destroy(mplp.bed);
     if (mplp.bed_itr) regitr_destroy(mplp.bed_itr);
     if (mplp.reg) regidx_destroy(mplp.reg);
+
+ err:
     bam_smpl_destroy(mplp.bsmpl);
 
     return ret;
