@@ -1,6 +1,6 @@
 /*  reheader.c -- reheader subcommand.
 
-    Copyright (C) 2014-2018 Genome Research Ltd.
+    Copyright (C) 2014-2022,2024 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -33,24 +33,177 @@ THE SOFTWARE.  */
 #include <inttypes.h>
 #include <fcntl.h>
 #include <math.h>
+#ifdef _WIN32
+#include <windows.h>
+#endif
 #include <htslib/vcf.h>
 #include <htslib/bgzf.h>
 #include <htslib/tbx.h> // for hts_get_bgzfp()
 #include <htslib/kseq.h>
 #include <htslib/thread_pool.h>
+#include <htslib/faidx.h>
+#include <htslib/khash_str2int.h>
 #include "bcftools.h"
 #include "khash_str2str.h"
 
 typedef struct _args_t
 {
     char **argv, *fname, *samples_fname, *header_fname, *output_fname;
+    char *fai_fname;
     htsFile *fp;
+    faidx_t *fai;
     htsFormat type;
     htsThreadPool *threads;
     int argc, n_threads;
 }
 args_t;
 
+static inline int is_escaped(const char *min, const char *str)
+{
+    int n = 0;
+    while ( --str>=min && *str=='\\' ) n++;
+    return n%2;
+}
+static char *copy_and_update_contig_line(faidx_t *fai, char *line, void *chr_seen, kstring_t *dst)
+{
+    kstring_t key = {0,0,0}, val = {0,0,0}, tmp = {0,0,0};
+    char *chr_name = NULL, *p, *q = line + 9;   // skip ##contig=
+    char *end = q;
+    int nopen = 1;
+    hts_pos_t chr_len = 0;
+    while ( *end && *end!='\n' ) end++;
+    while ( *q && *q!='\n' && nopen>0 )
+    {
+        p = ++q;
+        while ( *q && (*q==' ' || *q=='\t') ) { p++; q++; }
+        // ^[A-Za-z_][0-9A-Za-z_.]*$
+        if (p==q && *q && (isalpha(*q) || *q=='_'))
+        {
+            q++;
+            while ( *q && (isalnum(*q) || *q=='_' || *q=='.') ) q++;
+        }
+        int n = q-p;
+        int m = 0;
+        while ( *q && (*q==' ' || *q=='\t') ) { q++; m++; }
+        if ( *q!='=' || !n )
+        {
+            char *x = q;
+            while ( *x && *x!='\n' ) x++;
+            *x = '\0';
+            error("Could not parse the line: %s [%s][%s]\n", line,p,q);
+        }
+        key.l = 0;
+        kputsn(p,q-p-m,&key);
+        p = ++q;
+        while ( *q && (*q==' ' || *q=='\t') ) { p++; q++; }
+        int quoted = *p=='"' ? 1 : 0;
+        if ( quoted ) p++, q++;
+        while ( *q && *q != '\n' )
+        {
+            if ( quoted ) { if ( *q=='"' && !is_escaped(p,q) ) break; }
+            else
+            {
+                if ( *q=='<' ) nopen++;
+                if ( *q=='>' ) nopen--;
+                if ( !nopen ) break;
+                if ( *q==',' && nopen==1 ) break;
+            }
+            q++;
+        }
+        char *r = q;
+        while ( r > p && r[-1] == ' ' ) r--;
+        val.l = 0;
+        kputsn(p,r-p,&val);
+        if ( quoted && *q=='"' ) q++;
+        if ( *q=='>' ) { nopen--; q++; }
+        if ( !strcmp("length",key.s) ) continue;
+        if ( !strcmp("ID",key.s) )
+        {
+            if ( khash_str2int_has_key(chr_seen,val.s) ) continue;
+            chr_len = faidx_seq_len64(fai, val.s);
+            if ( chr_len==-1 )
+            {
+                free(val.s); free(key.s); free(tmp.s);
+                return end;   // the sequence is not in fai, remove
+            }
+            chr_name = strdup(val.s);
+            khash_str2int_inc(chr_seen, chr_name);
+            continue;
+        }
+        kputc(',',&tmp);
+        kputs(key.s,&tmp);
+        kputc('=',&tmp);
+        if ( quoted ) kputc('"',&tmp);
+        kputs(val.s,&tmp);
+        if ( quoted ) kputc('"',&tmp);
+    }
+    if ( !chr_name ) return end;
+    ksprintf(dst,"##contig=<ID=%s,length=%"PRIhts_pos"%s>",chr_name,chr_len,tmp.l ? tmp.s : "");
+    free(key.s); free(val.s); free(tmp.s);
+    return q;
+}
+char *init_tmp_prefix(const char *tmp_prefix)
+{
+    kstring_t prefix = {0,0,0};
+    if ( tmp_prefix )
+    {
+        ksprintf(&prefix,"%sXXXXXX",tmp_prefix);
+        return prefix.s;
+    }
+
+    char *tmpdir = getenv("TMPDIR");
+    if ( tmpdir )
+        kputs(tmpdir, &prefix);
+    else
+    {
+        #ifdef _WIN32
+            char tmp_path[MAX_PATH];
+            int ret = GetTempPath(MAX_PATH, tmp_path);
+            if (!ret || ret > MAX_PATH)
+                error("Could not get the path to the temporary folder\n");
+            kputs(tmp_path, &prefix);
+        #else
+            kputs("/tmp", &prefix);
+        #endif
+    }
+    kputs("/bcftools.XXXXXX", &prefix);
+    return prefix.s;
+}
+static void update_from_fai(faidx_t *fai, kstring_t *hdr_txt)
+{
+    kstring_t hdr_txt_new = {0,0,0};
+
+    // update the existing contig lines and remove lines not present in the fai file
+    void *chr_seen = khash_str2int_init();
+    char *tmp, *beg = hdr_txt->s;
+    while ( beg && *beg )
+    {
+        tmp = strstr(beg, "\n##contig=<");
+        if ( !tmp ) break;
+        kputsn(beg, tmp-beg+1, &hdr_txt_new);
+        size_t l_prev = hdr_txt_new.l;
+        beg = copy_and_update_contig_line(fai,tmp+1,chr_seen, &hdr_txt_new);
+        if ( l_prev==hdr_txt_new.l ) hdr_txt_new.l--;   // nothing was added, remove the newline
+    }
+    if ( !beg || !(tmp=strstr(beg,"\n#CHROM")) ) error("Failed to parse the header, #CHROM not found\n");
+    kputsn(beg, tmp-beg+1, &hdr_txt_new);
+
+    // add any new contig lines
+    int i, n = faidx_nseq(fai);
+    for (i=0; i<n; i++)
+    {
+        if ( khash_str2int_has_key(chr_seen,faidx_iseq(fai,i)) ) continue;
+        ksprintf(&hdr_txt_new,"##contig=<ID=%s,length=%"PRIhts_pos">\n",faidx_iseq(fai,i),faidx_seq_len64(fai,faidx_iseq(fai,i)));
+    }
+    kputs(tmp+1,&hdr_txt_new);
+
+    // Switch the new header content for the old
+    free(hdr_txt->s);
+    memcpy(hdr_txt, &hdr_txt_new, sizeof(*hdr_txt));
+
+    khash_str2int_destroy_free(chr_seen);
+}
+
 static void read_header_file(char *fname, kstring_t *hdr)
 {
     kstring_t tmp = {0,0,0};
@@ -115,8 +268,8 @@ static int set_sample_pairs(char **samples, int nsamples, kstring_t *hdr, int id
     hdr->s[hdr->l] = 0;
 
     kstring_t tmp = {0,0,0};
-    i = j = n = 0;
-    while ( hdr->s[idx+i] && hdr->s[idx+i])
+    i = j = n = 0;  // i:traverse the #CHROM line 1 by 1; j:points to the last column
+    while ( hdr->s[idx+i] )
     {
         if ( hdr->s[idx+i]=='\t' )
         {
@@ -124,8 +277,8 @@ static int set_sample_pairs(char **samples, int nsamples, kstring_t *hdr, int id
 
             if ( ++n>9 )
             {
-                char *ori = khash_str2str_get(hash,hdr->s+idx+j);
-                kputs(ori ? ori : hdr->s+idx+j, &tmp);
+                char *new_name = khash_str2str_get(hash,hdr->s+idx+j);
+                kputs(new_name ? new_name : hdr->s+idx+j, &tmp);
             }
             else
                 kputs(hdr->s+idx+j, &tmp);
@@ -137,8 +290,8 @@ static int set_sample_pairs(char **samples, int nsamples, kstring_t *hdr, int id
         }
         i++;
     }
-    char *ori = khash_str2str_get(hash,hdr->s+idx+j);
-    kputs(ori ? ori : hdr->s+idx+j, &tmp);
+    char *new_name = khash_str2str_get(hash,hdr->s+idx+j);
+    kputs(new_name ? new_name : hdr->s+idx+j, &tmp);
 
     khash_str2str_destroy_free_all(hash);
 
@@ -159,7 +312,13 @@ static void set_samples(char **samples, int nsamples, kstring_t *hdr)
         if ( hdr->s[i]=='\t' ) ncols++;
         i--;
     }
-    if ( i<0 || strncmp(hdr->s+i+1,"#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT",45) ) error("Could not parse the header: %s\n", hdr->s);
+    if ( i<0 || strncmp(hdr->s+i+1,"#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT",45) )
+    {
+        if ( i>0 && !strncmp(hdr->s+i+1,"#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO",38) )
+            error("Error: missing FORMAT fields, cowardly refusing to add samples\n");
+
+        error("Could not parse the header: %s\n", hdr->s);
+    }
 
     // Are the samples "old-sample new-sample" pairs?
     if ( set_sample_pairs(samples,nsamples,hdr, i+1) ) return;
@@ -230,12 +389,19 @@ static void reheader_vcf_gz(args_t *args)
     int nsamples = 0;
     char **samples = NULL;
     if ( args->samples_fname )
+    {
         samples = hts_readlines(args->samples_fname, &nsamples);
+        if ( !samples || !nsamples ) error("Error reading the --samples file \"%s\"\n", args->samples_fname);
+    }
     if ( args->header_fname )
     {
         free(hdr.s); hdr.s = NULL; hdr.l = hdr.m = 0;
         read_header_file(args->header_fname, &hdr);
     }
+
+    if ( args->fai )
+        update_from_fai(args->fai, &hdr);
+
     if ( samples )
     {
         set_samples(samples, nsamples, &hdr);
@@ -249,7 +415,7 @@ static void reheader_vcf_gz(args_t *args)
     if ( bgzf_write(bgzf_out, hdr.s, hdr.l) < 0 ) error("Can't write BGZF header (code %d)\n", bgzf_out->errcode);
     free(hdr.s);
 
-    // Output all remainig data read with the header block
+    // Output all remaining data read with the header block
     if ( fp->block_length - skip_until > 0 )
     {
         if ( bgzf_write(bgzf_out, buffer+skip_until, fp->block_length-skip_until)<0 ) error("Error: %d\n",fp->errcode);
@@ -286,12 +452,19 @@ static void reheader_vcf(args_t *args)
     int nsamples = 0;
     char **samples = NULL;
     if ( args->samples_fname )
+    {
         samples = hts_readlines(args->samples_fname, &nsamples);
+        if ( !samples || !nsamples ) error("Error reading the --samples file \"%s\"\n", args->samples_fname);
+    }
     if ( args->header_fname )
     {
         free(hdr.s); hdr.s = NULL; hdr.l = hdr.m = 0;
         read_header_file(args->header_fname, &hdr);
     }
+
+    if ( args->fai )
+        update_from_fai(args->fai, &hdr);
+
     if ( samples )
     {
         set_samples(samples, nsamples, &hdr);
@@ -313,8 +486,8 @@ static void reheader_vcf(args_t *args)
         kputc('\n',&fp->line);
         if ( write(out, fp->line.s, fp->line.l)!=fp->line.l ) error("Failed to write %"PRIu64" bytes\n", (uint64_t)fp->line.l);
     }
-    hts_close(fp);
-    close(out);
+    if ( hts_close(fp)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,args->fname);
+    if ( close(out)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,args->output_fname);
 }
 
 static bcf_hdr_t *strip_header(bcf_hdr_t *src, bcf_hdr_t *dst)
@@ -346,12 +519,14 @@ static bcf_hdr_t *strip_header(bcf_hdr_t *src, bcf_hdr_t *dst)
             if ( j>=0 )
             {
                 j = atoi(src_hrec->vals[j]);
-                hrec_add_idx(tmp, j);
+                if (hrec_add_idx(tmp, j) < 0)
+                    error_errno("[%s] Failed to add IDX header", __func__);
             }
             bcf_hdr_add_hrec(out, tmp);
         }
     }
-    bcf_hdr_sync(out);
+    if (bcf_hdr_sync(out) < 0)
+        error_errno("[%s] Failed to update header", __func__);
     for (i=0; i<dst->nhrec; i++)
     {
         // finally add new structured fields
@@ -375,11 +550,10 @@ static void reheader_bcf(args_t *args, int is_compressed)
 
     if ( args->n_threads > 0 )
     {
-        args->threads = calloc(1, sizeof(*args->threads));
+        args->threads = (htsThreadPool *) calloc(1, sizeof(htsThreadPool));
         if ( !args->threads ) error("Could not allocate memory\n");
         if ( !(args->threads->pool = hts_tpool_init(args->n_threads)) ) error("Could not initialize threading\n");
-        BGZF *bgzf = hts_get_bgzfp(fp);
-        if ( bgzf ) bgzf_thread_pool(bgzf, args->threads->pool, args->threads->qsize);
+        hts_set_thread_pool(fp, args->threads);
     }
 
     bcf_hdr_t *hdr = bcf_hdr_read(fp); if ( !hdr ) error("Failed to read the header: %s\n", args->fname);
@@ -389,12 +563,19 @@ static void reheader_bcf(args_t *args, int is_compressed)
     int i, nsamples = 0;
     char **samples = NULL;
     if ( args->samples_fname )
+    {
         samples = hts_readlines(args->samples_fname, &nsamples);
+        if ( !samples || !nsamples ) error("Error reading the --samples file \"%s\"\n", args->samples_fname);
+    }
     if ( args->header_fname )
     {
         free(htxt.s); htxt.s = NULL; htxt.l = htxt.m = 0;
         read_header_file(args->header_fname, &htxt);
     }
+
+    if ( args->fai )
+        update_from_fai(args->fai, &htxt);
+
     if ( samples )
     {
         set_samples(samples, nsamples, &htxt);
@@ -410,11 +591,8 @@ static void reheader_bcf(args_t *args, int is_compressed)
     htsFile *fp_out = hts_open(args->output_fname ? args->output_fname : "-",is_compressed ? "wb" : "wbu");
     if ( !fp_out ) error("%s: %s\n", args->output_fname ? args->output_fname : "-", strerror(errno));
     if ( args->threads )
-    {
-        BGZF *bgzf = hts_get_bgzfp(fp_out);
-        if ( bgzf ) bgzf_thread_pool(bgzf, args->threads->pool, args->threads->qsize);
-    }
-    bcf_hdr_write(fp_out, hdr_out);
+        hts_set_thread_pool(fp_out, args->threads);
+    if ( bcf_hdr_write(fp_out, hdr_out)!=0 ) error("[%s] Error: cannot write the header to %s\n", __func__,args->output_fname ? args->output_fname : "standard output");
 
     bcf1_t *rec = bcf_init();
     while ( bcf_read(fp, hdr, rec)==0 )
@@ -459,13 +637,13 @@ static void reheader_bcf(args_t *args, int is_compressed)
         if ( i!=rec->n_fmt )
             error("The FORMAT tag is not defined: \"%s\"\n", bcf_hdr_int2id(hdr,BCF_DT_ID,rec->d.fmt[i].id));
 
-        bcf_write(fp_out,hdr_out,rec);
+        if ( bcf_write(fp_out,hdr_out,rec)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname ? args->output_fname : "standard output");
     }
     bcf_destroy(rec);
 
     free(htxt.s);
-    hts_close(fp_out);
-    hts_close(fp);
+    if ( hts_close(fp_out)!=0 ) error("[%s] Error: failed to close the file %s\n",__func__,args->output_fname ? args->output_fname : "standard output");
+    if ( hts_close(fp)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,args->fname);
     bcf_hdr_destroy(hdr_out);
     bcf_hdr_destroy(hdr);
     if ( args->threads )
@@ -483,10 +661,22 @@ static void usage(args_t *args)
     fprintf(stderr, "Usage:   bcftools reheader [OPTIONS] <in.vcf.gz>\n");
     fprintf(stderr, "\n");
     fprintf(stderr, "Options:\n");
-    fprintf(stderr, "    -h, --header <file>     new header\n");
-    fprintf(stderr, "    -o, --output <file>     write output to a file [standard output]\n");
-    fprintf(stderr, "    -s, --samples <file>    new sample names\n");
-    fprintf(stderr, "        --threads <int>     number of extra compression threads (BCF only) [0]\n");
+    fprintf(stderr, "    -f, --fai FILE             update sequences and their lengths from the .fai file\n");
+    fprintf(stderr, "    -h, --header FILE          new header\n");
+    fprintf(stderr, "    -o, --output FILE          write output to a file [standard output]\n");
+    fprintf(stderr, "    -s, --samples FILE         new sample names\n");
+    fprintf(stderr, "    -T, --temp-prefix PATH     ignored; was template for temporary file name\n");
+    fprintf(stderr, "        --threads INT          use multithreading with <int> worker threads (BCF only) [0]\n");
+    fprintf(stderr, "\n");
+    fprintf(stderr, "Example:\n");
+    fprintf(stderr, "   # Write out the header to be modified\n");
+    fprintf(stderr, "   bcftools view -h old.bcf > header.txt\n");
+    fprintf(stderr, "\n");
+    fprintf(stderr, "   # Edit the header using your favorite text editor\n");
+    fprintf(stderr, "   vi header.txt\n");
+    fprintf(stderr, "\n");
+    fprintf(stderr, "   # Reheader the file\n");
+    fprintf(stderr, "   bcftools reheader -h header.txt -o new.bcf old.bcf\n");
     fprintf(stderr, "\n");
     exit(1);
 }
@@ -496,24 +686,28 @@ int main_reheader(int argc, char *argv[])
     int c;
     args_t *args  = (args_t*) calloc(1,sizeof(args_t));
     args->argc    = argc; args->argv = argv;
-    
+
     static struct option loptions[] =
     {
+        {"temp-prefix",1,0,'T'},
+        {"fai",1,0,'f'},
         {"output",1,0,'o'},
         {"header",1,0,'h'},
         {"samples",1,0,'s'},
         {"threads",1,NULL,1},
         {0,0,0,0}
     };
-    while ((c = getopt_long(argc, argv, "s:h:o:",loptions,NULL)) >= 0)
+    while ((c = getopt_long(argc, argv, "s:h:o:f:T:",loptions,NULL)) >= 0)
     {
         switch (c)
         {
             case  1 : args->n_threads = strtol(optarg, 0, 0); break;
+            case 'T': break; // unused - was temp file prefix
+            case 'f': args->fai_fname = optarg; break;
             case 'o': args->output_fname = optarg; break;
             case 's': args->samples_fname = optarg; break;
             case 'h': args->header_fname = optarg; break;
-            case '?': usage(args);
+            case '?': usage(args); break;
             default: error("Unknown argument: %s\n", optarg);
         }
     }
@@ -525,23 +719,33 @@ int main_reheader(int argc, char *argv[])
     }
     else args->fname = argv[optind];
 
-    if ( !args->samples_fname && !args->header_fname ) usage(args);
+    if ( args->fai_fname ) {
+        args->fai = fai_load3(args->fai_fname,args->fai_fname,NULL,FAI_FASTA);
+        if ( !args->fai ) error("Could not parse %s\n", args->fai_fname);
+    }
+    if ( !args->samples_fname && !args->header_fname && !args->fai) usage(args);
     if ( !args->fname ) usage(args);
 
     args->fp = hts_open(args->fname,"r");
-    if ( !args->fp ) error("Failed to open: %s\n", args->fname);
+    if ( !args->fp ) error("Failed to read from %s\n", !strcmp("-",args->fname)?"standard input":args->fname);
     args->type = *hts_get_format(args->fp);
 
     if ( args->type.format==vcf )
     {
-        if ( args->type.compression==bgzf || args->type.compression==gzip )
+        if ( args->type.compression==bgzf )
             reheader_vcf_gz(args);
-        else
+        else if ( args->type.compression==no_compression )
             reheader_vcf(args);
+        else if ( args->type.compression==gzip )
+            error("Error: cannot reheader gzip-compressed files, first convert with `bcftools view --output-type` to a supported format\n");
+        else
+            error("Error: the compression type of \"%s\" is not recognised/supported\n", args->fname);
     }
     else
         reheader_bcf(args, args->type.compression==bgzf || args->type.compression==gzip);
 
+    if (args->fai)
+        fai_destroy(args->fai);
     free(args);
     return 0;
 }
