@@ -8,6 +8,7 @@
 
 import fnmatch
 import os
+from pathlib import Path
 import re
 import itertools
 import shutil
@@ -35,8 +36,8 @@ EXCLUDE = {
         "peakfit.h",
         "polysomy.c"),
     "htslib": (
-        'htslib/tabix.c', 'htslib/bgzip.c',
-        'htslib/htsfile.c',
+        'annot-tsv.c', 'bgzip.c', 'htsfile.c', 'tabix.c',
+        'hts_probe_cc.sh',
         "samples", "test", "tests"),
 }
 
@@ -119,6 +120,7 @@ def _update_pysam_files(cf, destdir):
                 if fn == "bamtk.c":
                     lines = re.sub(r'(#include "version.h")', r'\1\n#include "samtools_config_vars.h"', lines)
                     lines = re.sub(r'(else if.*"tview")', r'//\1', lines)
+                    lines = re.sub(r'(if[ (]*fclose)', r'if (0) { //\1', lines)
 
                 outfile.write(lines)
 
@@ -152,9 +154,10 @@ if len(sys.argv) >= 1:
                              locate("version.sh", srcdir, exclude_htslib=True))
 
     if dest == "htslib":
-        # Add build files, including *.ac *.in *.mk *.m4 *.sh
+        # Add build files, including *.ac config.{guess,sub} *.in *.mk *.m4 *.sh
         mfiles = itertools.chain(mfiles, locate("Makefile", srcdir),
                                  locate("configure", srcdir),
+                                 locate("config.*", srcdir),
                                  locate("*.[aims][cnk4h]", srcdir, exclude))
 
     ncopied = 0
@@ -243,6 +246,9 @@ if len(sys.argv) >= 1:
     _update_version_file(C_VERSION[dest], version + " (pysam)", "pysam/version.h")
     _update_version_doc_file(dest, version, "README.rst")
     _update_version_doc_file(dest, version, "doc/index.rst")
+    if dest in MAIN:
+        path = Path("doc/conf.py")
+        path.write_text(re.sub(rf"doc/[\d.]*/{dest}", rf"doc/{version}/{dest}", path.read_text()))
 
     sys.exit(0)
 
