@@ -1,7 +1,7 @@
 /*  bam2depth.c -- depth subcommand.
 
     Copyright (C) 2011, 2012 Broad Institute.
-    Copyright (C) 2012-2016, 2018, 2019-2022 Genome Research Ltd.
+    Copyright (C) 2012-2016, 2018, 2019-2022, 2024 Genome Research Ltd.
 
     Author: Heng Li <lh3@sanger.ac.uk> (to 2020)
     Author: James Bonfield <jkb@sanger.ac.uk> (2021 rewrite)
@@ -72,6 +72,8 @@ typedef struct {
 typedef struct {
     int header;
     int flag;
+    int incl_flag;
+    int require_flag;
     int min_qual;
     int min_mqual;
     int min_len;
@@ -156,6 +158,42 @@ hts_pos_t qlen_used(bam1_t *b) {
 
 }
 
+// Without HTS_OPT3 gcc11 and earlier don't vectorize this, nor even
+// unroll it.  By adding HTS_OPT3 we can force a better level of optimization.
+// On an Illumina BAM with gcc11 -O2, with HTS_OPT3 is 9% quicker for the
+// entire process (ie decompress, iterator, aggregate & report)
+static inline void HTS_OPT3 incr_hist(int *hist, int oplen) {
+    const int N = 16;
+    int k;
+    for (k = 0; k < (oplen & ~(N-1)); k+=N) {
+        for (int i = 0; i < N; i++)
+            hist[k+i]++;
+    }
+    for (; k < oplen; k++)
+        hist[k]++;
+}
+
+static inline void HTS_OPT3
+incr_hist_qual(int *hist, uint8_t *qual, int min_qual, int oplen) {
+    if (!min_qual) {
+        incr_hist(hist, oplen);
+        return;
+    }
+
+    int k;
+    for (k = 0; k < (oplen & ~31); k+=32) {
+        int pass[32];
+        // Two separate loops helps clang to vectorize
+        for (int i = 0; i < 32; i++)
+            pass[i]=qual[k+i]>=min_qual;
+
+        for (int i = 0; i < 32; i++)
+            hist[k+i]+=pass[i];
+    }
+    for (; k < oplen; k++)
+        hist[k]+=qual[k]>=min_qual;
+}
+
 // Adds the depth for a single read to a depth_hist struct.
 // For just one file, this is easy.  We just have a circular buffer
 // where we increment values for bits that overlap existing data
@@ -377,7 +415,6 @@ static int add_depth(depth_opt *opt, depth_hist *dh, sam_hdr_t *h, bam1_t *b,
                 // We've explicitly asked to include them, and the quality
                 // is wrong anyway (it's the neighbouring base).  We do this
                 // for now for compatibility with the old depth command.
-
                 if (spos < b->core.l_qseq)
                     for (; k < oplen; k++, i++)
                         hist[i & hmask]+=qual[spos]>=min_qual;
@@ -389,80 +426,30 @@ static int add_depth(depth_opt *opt, depth_hist *dh, sam_hdr_t *h, bam1_t *b,
 
         case BAM_CMATCH:
         case BAM_CEQUAL:
-        case BAM_CDIFF:
-            if ((i & hmask) < ((i+oplen) & hmask)) {
-                // Optimisation when not wrapping around
-
-                // Unrolling doesn't help clang, but helps gcc,
-                // especially when not using -O3.
-                int *hist = &dh->hist[file][i & hmask];
-                if (min_qual || overlap_clip) {
-                    k = 0;
-                    if (overlap_clip) {
-                        if (i+oplen < overlap_clip) {
-                            i += oplen;
-                            spos += oplen;
-                            break;
-                        } else if (i < overlap_clip) {
-                            oplen -= overlap_clip - i;
-                            spos += overlap_clip - i;
-                            hist += overlap_clip - i;
-                            i = overlap_clip;
-                        }
-                    }
-
-                    // approx 50% of this func cpu time in this loop
-                    for (; k < (oplen & ~7); k+=8) {
-                        hist[k+0]+=qual[spos+0]>=min_qual;
-                        hist[k+1]+=qual[spos+1]>=min_qual;
-                        hist[k+2]+=qual[spos+2]>=min_qual;
-                        hist[k+3]+=qual[spos+3]>=min_qual;
-                        hist[k+4]+=qual[spos+4]>=min_qual;
-                        hist[k+5]+=qual[spos+5]>=min_qual;
-                        hist[k+6]+=qual[spos+6]>=min_qual;
-                        hist[k+7]+=qual[spos+7]>=min_qual;
-                        spos += 8;
-                    }
-                } else {
-                    // easier to vectorize when no min_qual
-                    for (k = 0; k < (oplen & ~7); k+=8) {
-                        hist[k+0]++;
-                        hist[k+1]++;
-                        hist[k+2]++;
-                        hist[k+3]++;
-                        hist[k+4]++;
-                        hist[k+5]++;
-                        hist[k+6]++;
-                        hist[k+7]++;
-                    }
-                    spos += k;
+        case BAM_CDIFF: {
+            int *hist = dh->hist[file];
+            k = 0;
+            if (overlap_clip) {
+                if (i+oplen < overlap_clip) {
+                    i += oplen;
+                    break;
+                } else if (i < overlap_clip) {
+                    oplen -= overlap_clip - i;
+                    spos += overlap_clip - i;
+                    i = overlap_clip;
                 }
-                for (; k < oplen && spos < b->core.l_qseq; k++, spos++)
-                    hist[k]+=qual[spos]>=min_qual;
-                for (; k < oplen; k++, spos++)
-                    hist[k]++;
-                i += oplen;
-            } else {
-                // Simple to understand case, but slower.
-                // We use this only for reads with wrap-around.
-                int *hist = dh->hist[file];
-                k = 0;
-                if (overlap_clip) {
-                    if (i+oplen < overlap_clip) {
-                        i += oplen;
-                        break;
-                    } else if (i < overlap_clip) {
-                        oplen -= overlap_clip - i;
-                        spos += overlap_clip - i;
-                        i = overlap_clip;
-                    }
-                }
-                for (; k < oplen && spos < b->core.l_qseq; k++, i++, spos++)
-                    hist[i & hmask]+=qual[spos]>=min_qual;
-                for (; k < oplen; k++, i++, spos++)
-                    hist[i & hmask]++;
             }
+
+            int len = (i & hmask) < ((i+oplen) & hmask)
+                ? oplen                  // doesn't wrap around
+                : dh->size - (i&hmask);  // does wrap around
+            incr_hist_qual(&hist[i & hmask], &qual[spos], min_qual, len);
+            if (oplen > len)
+                incr_hist_qual(hist, &qual[spos+len], min_qual, oplen-len);
+            spos += oplen;
+            i += oplen;
             break;
+        }
 
         case BAM_CINS:
         case BAM_CSOFT_CLIP:
@@ -564,7 +551,11 @@ static int fastdepth_core(depth_opt *opt, uint32_t nfiles, char **fn,
             if (b[i]->core.tid < 0)
                 continue;
             if (b[i]->core.flag & opt->flag)
-                continue;
+                continue; // must have none of the flags set
+            if (opt->incl_flag && (b[i]->core.flag & opt->incl_flag) == 0)
+                continue; // must have at least one flag set
+            if ((b[i]->core.flag & opt->require_flag) != opt->require_flag)
+                continue; // must have all lags set
             if (b[i]->core.qual < opt->min_mqual)
                 continue;
 
@@ -654,7 +645,11 @@ static int fastdepth_core(depth_opt *opt, uint32_t nfiles, char **fn,
             if (b[i]->core.tid < 0)
                 continue;
             if (b[i]->core.flag & opt->flag)
-                continue;
+                continue; // must have none of the flags set
+            if (opt->incl_flag && (b[i]->core.flag & opt->incl_flag) == 0)
+                continue; // must have at least one flag set
+            if ((b[i]->core.flag & opt->require_flag) != opt->require_flag)
+                continue; // must have all lags set
             if (b[i]->core.qual < opt->min_mqual)
                 continue;
 
@@ -712,8 +707,14 @@ static void usage_exit(FILE *fp, int exit_status)
     fprintf(fp, "  -b FILE      Use bed FILE for list of regions\n");
     fprintf(fp, "  -f FILE      Specify list of input BAM/SAM/CRAM filenames\n");
     fprintf(fp, "  -X           Use custom index files (in -X *.bam *.bam.bai order)\n");
-    fprintf(fp, "  -g INT       Remove specified flags from default flag filter\n");
-    fprintf(fp, "  -G INT       Add specified flags to the default flag filter\n");
+    fprintf(fp, "  -g INT       Remove specified flags from default filter-out flag list\n");
+    fprintf(fp, "  -G, --excl-flags FLAGS\n");
+    fprintf(fp, "               Add specified flags to the  default filter-out flag list\n");
+    fprintf(fp, "               [UNMAP,SECONDARY,QCFAIL,DUP]\n");
+    fprintf(fp, "      --incl-flags FLAGS\n");
+    fprintf(fp, "               Only include records with at least one the FLAGs present [0]\n");
+    fprintf(fp, "      --require-flags FLAGS\n");
+    fprintf(fp, "               Only include records with all of the FLAGs present [0]\n");
     fprintf(fp, "  -H           Print a file header line\n");
     fprintf(fp, "  -l INT       Minimum read length [0]\n");
     fprintf(fp, "  -o FILE      Write output to FILE [stdout]\n");
@@ -721,7 +722,6 @@ static void usage_exit(FILE *fp, int exit_status)
                 "               Filter bases with base quality smaller than INT [0]\n");
     fprintf(fp, "  -Q, --min-MQ INT\n"
                 "               Filter alignments with mapping quality smaller than INT [0]\n");
-    fprintf(fp, "  -H           Print a file header\n");
     fprintf(fp, "  -J           Include reads with deletions in depth computation\n");
     fprintf(fp, "  -s           Do not count overlapping reads within a template\n");
     sam_global_opt_help(fp, "-.--.@-.");
@@ -730,7 +730,7 @@ static void usage_exit(FILE *fp, int exit_status)
 
 int main_depth(int argc, char *argv[])
 {
-    int nfiles, i;
+    int nfiles, i, tmp_flag;
     samFile **fp;
     sam_hdr_t **header;
     int c, has_index_file = 0;
@@ -738,6 +738,8 @@ int main_depth(int argc, char *argv[])
     char *out_file = NULL;
     depth_opt opt = {
         .flag = BAM_FUNMAP | BAM_FSECONDARY | BAM_FDUP | BAM_FQCFAIL,
+        .incl_flag = 0,
+        .require_flag = 0,
         .min_qual = 0,
         .min_mqual = 0,
         .skip_del = 1,
@@ -752,10 +754,13 @@ int main_depth(int argc, char *argv[])
 
     sam_global_args ga = SAM_GLOBAL_ARGS_INIT;
     static const struct option lopts[] = {
-        {"min-MQ", required_argument, NULL, 'Q'},
-        {"min-mq", required_argument, NULL, 'Q'},
-        {"min-BQ", required_argument, NULL, 'q'},
-        {"min-bq", required_argument, NULL, 'q'},
+        {"min-MQ",        required_argument, NULL, 'Q'},
+        {"min-mq",        required_argument, NULL, 'Q'},
+        {"min-BQ",        required_argument, NULL, 'q'},
+        {"min-bq",        required_argument, NULL, 'q'},
+        {"excl-flags",    required_argument, NULL, 'G'},
+        {"incl-flags",    required_argument, NULL, 1},
+        {"require-flags", required_argument, NULL, 2},
         SAM_OPT_GLOBAL_OPTIONS('-', 0, '-', '-', 0, '@'),
         {NULL, 0, NULL, 0}
     };
@@ -786,10 +791,44 @@ int main_depth(int argc, char *argv[])
             break;
 
         case 'g':
-            opt.flag &= ~bam_str2flag(optarg);
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("depth", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            opt.flag &= ~tmp_flag;
             break;
-        case 'G':
-            opt.flag |= bam_str2flag(optarg);
+        case 'G': // reject if any set
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("depth", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            opt.flag |= tmp_flag;
+            break;
+        case 1: // reject unless at least one set (0 means ignore option)
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("depth", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            opt.incl_flag |= tmp_flag;
+            break;
+        case 2: // reject unless all set
+            tmp_flag = bam_str2flag(optarg);
+
+            if (tmp_flag < 0) {
+                print_error("depth", "Unknown flag '%s'", optarg);
+                return 1;
+            }
+
+            opt.require_flag |= tmp_flag;
             break;
 
         case 'l':
