@@ -1,7 +1,7 @@
 /*  bam_ampliconclip.c -- loads amplicon primers from a BED file and cuts reads
                           from the 5' end.
 
-    Copyright (C) 2020-2021 Genome Research Ltd.
+    Copyright (C) 2020-2024 Genome Research Ltd.
 
     Authors: Andrew Whitwham <aw7@sanger.ac.uk>
              Rob Davies <rmd+git@sanger.ac.uk>
@@ -59,8 +59,10 @@ typedef struct {
     int oa_tag;
     int del_tag;
     int tol;
+    int unmap_len;
     char *arg_list;
     char *stats_file;
+    char *primer_counts_file;
     char *rejects_file;
 } cl_param_t;
 
@@ -71,49 +73,63 @@ static int bed_entry_sort(const void *av, const void *bv) {
     return a->right < b->right ? -1 : (a->right == b->right ? 0 : 1);
 }
 
-
-int load_bed_file_multi_ref(char *infile, int get_strand, int sort_by_pos, khash_t(bed_list_hash) *bed_lists) {
+int load_bed_file_multi_ref(char *infile, int get_strand, int sort_by_pos, khash_t(bed_list_hash) *bed_lists, char ***ref_list, size_t *num_refs_out) {
     hFILE *fp;
     int line_count = 0, ret;
+
+    //variables to store the bed file data for each record
+    char ref[1024] = "";
     int64_t left, right;
+    char name[1024] = "", score[1024] = "";
+    char strand;
+
+    //hash table to store clipping results and bed file data
     kstring_t line = KS_INITIALIZE;
     bed_entry_list_t *list;
     khiter_t bed_itr;
 
+    //ordered ref names list
+    size_t ref_list_sz = 0;
+    size_t num_refs = 0;
+
+    if (ref_list)
+        *ref_list = NULL;
+
+    if (num_refs_out)
+        *num_refs_out = 0;
+
     if ((fp = hopen(infile, "r")) == NULL) {
         print_error_errno("amplicon", "unable to open file %s.", infile);
         return 1;
     }
 
-    char ref[1024];
 
     while (line.l = 0, kgetline(&line, (kgets_func *)hgets, fp) >= 0) {
         line_count++;
         int hret;
-        char strand;
 
         if (line.l == 0 || *line.s == '#') continue;
         if (strncmp(line.s, "track ", 6) == 0) continue;
         if (strncmp(line.s, "browser ", 8) == 0) continue;
 
-        if (get_strand) {
-            if (sscanf(line.s, "%1023s %"SCNd64" %"SCNd64" %*s %*s %c",
-                       ref, &left, &right, &strand) != 4) {
-                fprintf(stderr, "[amplicon] error: bad bed file format in line %d of %s.\n"
-                                "(N.B. ref/chrom name limited to 1023 characters.)\n",
-                                    line_count, infile);
-                ret = 1;
-                goto error;
-            }
-        } else {
-            if (sscanf(line.s, "%1023s %"SCNd64" %"SCNd64,
-                       ref, &left, &right) != 3) {
-                fprintf(stderr, "[amplicon] error: bad bed file format in line %d of %s\n"
-                                "(N.B. ref/chrom name limited to 1023 characters.)\n",
-                                    line_count, infile);
-                ret = 1;
-                goto error;
-            }
+        // A list of the maximal number of columns we may want to parse.
+        // There may be more, but we don't use the ones beyond these.
+        const char *const scanf_str =
+            "%1023s %"SCNd64" %"SCNd64" %1023s %1023s %c";
+
+        // Extract the data from the line into the variables.
+        // Variables corresponding to any missing columns will remain
+        // uninitialised.  We asked for all columns, but cols_parsed will
+        // return how many we found which can be validated against num_columns.
+        int cols_parsed = sscanf(line.s, scanf_str, ref,
+                                 &left, &right, name, score, &strand);
+        if (cols_parsed < (get_strand ? 6 : 3)) {
+            fprintf(stderr, "[amplicon] error: invalid bed file format in line %d of %s.\n"
+                    "Parsed %d columns, but need at least %d\n"
+                    "(N.B. ref/chrom name limited to 1023 characters.)\n",
+                    line_count, infile, cols_parsed, get_strand ? 6 : 3);
+            ret = 1;
+            goto error;
         }
 
         bed_itr = kh_get(bed_list_hash, bed_lists, ref);
@@ -127,6 +143,14 @@ int load_bed_file_multi_ref(char *infile, int get_strand, int sort_by_pos, khash
                 goto error;
             }
 
+            if (ref_list) {
+                if (hts_resize(char **, num_refs + 1, &ref_list_sz, ref_list, 0) < 0) {
+                    fprintf(stderr, "[amplicon] error: unable to allocate memory for ref name list.\n");
+                    ret = 1;
+                    goto error;
+                }
+                (*ref_list)[num_refs++] = ref_name;
+            }
             bed_itr = kh_put(bed_list_hash, bed_lists, ref_name, &hret);
 
             if (hret > 0) {
@@ -146,6 +170,7 @@ int load_bed_file_multi_ref(char *infile, int get_strand, int sort_by_pos, khash
             list = &kh_val(bed_lists, bed_itr);
         }
 
+        // add the bed entry to the list, growing the list if necessary
         if (list->length == list->size) {
            bed_entry_t *tmp;
 
@@ -162,6 +187,26 @@ int load_bed_file_multi_ref(char *infile, int get_strand, int sort_by_pos, khash
 
         list->bp[list->length].left  = left;
         list->bp[list->length].right = right;
+        list->bp[list->length].name  = NULL;
+        list->bp[list->length].score = NULL;
+        if (cols_parsed >= 4) {
+            list->bp[list->length].name = strdup(name);
+            if (list->bp[list->length].name == NULL) {
+                fprintf(stderr, "[amplicon] error: unable to allocate memory for name in line %d of %s: %s.\n", line_count, infile, line.s);
+                ret = 1;
+                goto error;
+            }
+        }
+        if (cols_parsed >= 5) {
+            list->bp[list->length].score = strdup(score);
+            if (list->bp[list->length].score == NULL) {
+                fprintf(stderr, "[amplicon] error: unable to allocate memory for score in line %d of %s: %s\n", line_count, infile, line.s);
+                ret = 1;
+                goto error;
+            }
+        }
+
+        list->bp[list->length].num_reads = 0;
 
         if (get_strand) {
             if (strand == '+') {
@@ -197,6 +242,9 @@ int load_bed_file_multi_ref(char *infile, int get_strand, int sort_by_pos, khash
         ret = 1;
     }
 
+    if (num_refs_out)
+        *num_refs_out = num_refs;
+
 error:
     ks_free(&line);
 
@@ -213,7 +261,12 @@ void destroy_bed_hash(khash_t(bed_list_hash) *hash) {
 
     for (itr = kh_begin(hash); itr != kh_end(hash); ++itr) {
        if (kh_exist(hash, itr)) {
-           free(kh_val(hash, itr).bp);
+           bed_entry_list_t list = kh_val(hash, itr);
+           for (int i = 0; i < list.length; i++) {
+               free(list.bp[i].name);
+               free(list.bp[i].score);
+           }
+           free(list.bp);
            free((char *)kh_key(hash, itr));
            kh_key(hash, itr) = NULL;
         }
@@ -226,7 +279,7 @@ void destroy_bed_hash(khash_t(bed_list_hash) *hash) {
 static int matching_clip_site(bed_entry_list_t *sites, hts_pos_t pos,
                               int is_rev, int use_strand, int64_t longest,
                               cl_param_t *param) {
-    int i, size;  // may need this to be variable
+    int i, size, used_i;
     int tol = param->tol;
     int l = 0, mid = sites->length / 2, r = sites->length;
     int pos_tol = is_rev ? (pos > tol ? pos - tol : 0) : pos;
@@ -241,6 +294,7 @@ static int matching_clip_site(bed_entry_list_t *sites, hts_pos_t pos,
     }
 
     size = 0;
+    used_i = -1;
 
     for (i = l; i < sites->length; i++) {
         hts_pos_t mod_left, mod_right;
@@ -267,15 +321,19 @@ static int matching_clip_site(bed_entry_list_t *sites, hts_pos_t pos,
             if (is_rev) {
                 if (size < pos - sites->bp[i].left) {
                     size = pos - sites->bp[i].left;
+                    used_i = i;
                 }
             } else {
                 if (size < sites->bp[i].right - pos) {
                     size = sites->bp[i].right - pos;
+                    used_i = i;
                 }
             }
         }
     }
-
+    if (used_i >= 0) {
+        sites->bp[used_i].num_reads++;
+    }
     return size;
 }
 
@@ -308,18 +366,6 @@ static int bam_trim_left(bam1_t *rec, bam1_t *rec_out, uint32_t bases,
     memcpy(&rec_out->core, &rec->core, sizeof(rec->core));
     memcpy(rec_out->data, rec->data, rec->core.l_qname);
 
-    if (clipping == hard_clip && bases >= rec->core.l_qseq) {
-        rec_out->core.l_qseq = 0;
-        rec_out->core.n_cigar = 0;
-
-        if (orig_l_aux)
-            memcpy(bam_get_aux(rec_out), orig_aux, orig_l_aux);
-
-        rec_out->l_data = bam_get_aux(rec_out) - rec_out->data + orig_l_aux;
-
-        return 0;
-    }
-
     // Modify CIGAR
     new_cigar = bam_get_cigar(rec_out);
 
@@ -355,6 +401,19 @@ static int bam_trim_left(bam1_t *rec, bam1_t *rec_out, uint32_t bases,
             qry_removed += ref_remove;
         }
     } else {
+        if (clipping == hard_clip) {
+
+            rec_out->core.l_qseq = 0;
+            rec_out->core.n_cigar = 0;
+
+            if (orig_l_aux)
+                memcpy(bam_get_aux(rec_out), orig_aux, orig_l_aux);
+
+            rec_out->l_data = bam_get_aux(rec_out) - rec_out->data + orig_l_aux;
+
+            return 0;
+        }
+
         qry_removed = rec->core.l_qseq;
     }
 
@@ -409,7 +468,7 @@ static int bam_trim_left(bam1_t *rec, bam1_t *rec_out, uint32_t bases,
     }
 
     // Copy remaining QUAL
-    memmove(new_qual, orig_qual, rec->core.l_qseq - qry_removed);
+    memmove(new_qual, orig_qual + qry_removed, rec->core.l_qseq - qry_removed);
 
     // Set new l_qseq
     rec_out->core.l_qseq -= qry_removed;
@@ -457,17 +516,6 @@ static int bam_trim_right(bam1_t *rec, bam1_t *rec_out, uint32_t bases,
     memcpy(&rec_out->core, &rec->core, sizeof(rec->core));
     memcpy(rec_out->data, rec->data, rec->core.l_qname);
 
-    if (clipping == hard_clip && bases >= rec->core.l_qseq) {
-        rec_out->core.l_qseq = 0;
-        rec_out->core.n_cigar = 0;
-
-        if (orig_l_aux)
-            memcpy(bam_get_aux(rec_out), orig_aux, orig_l_aux);
-
-        rec_out->l_data = bam_get_aux(rec_out) - rec_out->data + orig_l_aux;
-        return 0;
-    }
-
     // Modify CIGAR here
     new_cigar = bam_get_cigar(rec_out);
 
@@ -500,6 +548,19 @@ static int bam_trim_right(bam1_t *rec, bam1_t *rec_out, uint32_t bases,
         if (qry_removed > 0) j++;
         if (hardclip > 0 && (clipping == soft_clip || qry_removed == 0)) j++;
     } else {
+        if (clipping == hard_clip) {
+
+            rec_out->core.l_qseq = 0;
+            rec_out->core.n_cigar = 0;
+
+            if (orig_l_aux)
+                memcpy(bam_get_aux(rec_out), orig_aux, orig_l_aux);
+
+            rec_out->l_data = bam_get_aux(rec_out) - rec_out->data + orig_l_aux;
+
+            return 0;
+        }
+
         qry_removed = rec->core.l_qseq;
         j = 0;
         if (hardclip > 0 && clipping == soft_clip) j++;
@@ -635,11 +696,16 @@ static int bam_clip(samFile *in, samFile *out, samFile *reject, char *bedfile,
     long filtered = 0, written = 0, failed = 0;
     kstring_t str = KS_INITIALIZE;
     kstring_t oat = KS_INITIALIZE;
+    kstring_t seq = KS_INITIALIZE;
     bed_entry_list_t *sites;
     FILE *stats_fp = stderr;
+    FILE *bed_count_summary_fp = stderr;
     khash_t(bed_list_hash) *bed_hash = kh_init(bed_list_hash);
+    char **bed_ref_list = NULL;
+    size_t num_bed_refs = 0;
 
-    if (load_bed_file_multi_ref(bedfile, param->use_strand, 1, bed_hash)) {
+    if (load_bed_file_multi_ref(bedfile, param->use_strand, 1, bed_hash,
+                                &bed_ref_list, &num_bed_refs)) {
         fprintf(stderr, "[ampliconclip] error: unable to load bed file.\n");
         goto fail;
     }
@@ -826,16 +892,46 @@ static int bam_clip(samFile *in, samFile *out, samFile *reject, char *bedfile,
                 }
             }
 
-            if (param->fail_len >= 0 || param->filter_len >= 0) {
-               hts_pos_t aql = active_query_len(b);
+            if (param->fail_len >= 0 || param->filter_len >= 0 || param->unmap_len >= 0) {
+                hts_pos_t aql = active_query_len(b);
 
-               if (param->fail_len >= 0 && aql <= param->fail_len) {
-                   b->core.flag |= BAM_FQCFAIL;
-               }
+                if (param->fail_len >= 0 && aql <= param->fail_len) {
+                    b->core.flag |= BAM_FQCFAIL;
+                }
 
-               if (param->filter_len >= 0 && aql <= param->filter_len) {
-                   filter = 1;
-               }
+                if (param->filter_len >= 0 && aql <= param->filter_len) {
+                    filter = 1;
+                }
+
+                if (param->unmap_len >= 0 && aql <= param->unmap_len) {
+
+                    if (ks_resize(&seq, b->core.l_qseq) < 0) {
+                        fprintf(stderr, "[ampliconclip] error: allocate memory for sequence %s\n", bam_get_seq(b));
+                        goto fail;
+                    }
+
+                    ks_clear(&seq);
+                    char *sb = ks_str(&seq);
+                    uint8_t *sequence = bam_get_seq(b);
+                    int i;
+
+                    for (i = 0; i < b->core.l_qseq ; ++i) {
+                        *sb++ = seq_nt16_str[bam_seqi(sequence, i)];
+                    }
+
+                    if (bam_set1(b_tmp, b->core.l_qname - b->core.l_extranul - 1, bam_get_qname(b),
+                                 (b->core.flag | BAM_FUNMAP), b->core.tid, b->core.pos, 0,
+                                 0, NULL, b->core.mtid, b->core.mpos, b->core.isize,
+                                 b->core.l_qseq, seq.s, (const char *)bam_get_qual(b),
+                                 bam_get_l_aux(b)) < 0) {
+                        fprintf(stderr, "[ampliconclip] error: could not unmap read %s\n", bam_get_seq(b));
+                        goto fail;
+                    }
+
+                    memcpy(bam_get_aux(b_tmp), bam_get_aux(b), bam_get_l_aux(b));
+                    b_tmp->l_data += bam_get_l_aux(b);
+                    swap_bams(&b, &b_tmp);
+                }
            }
 
            if (b->core.flag & BAM_FQCFAIL) {
@@ -903,13 +999,46 @@ static int bam_clip(samFile *in, samFile *out, samFile *reject, char *bedfile,
 
     if (file_open) {
         fclose(stats_fp);
+        file_open = 0;
+    }
+
+    if (param->primer_counts_file) {
+        if ((bed_count_summary_fp = fopen(param->primer_counts_file, "w")) == NULL) {
+            fprintf(stderr, "[ampliconclip] warning: cannot write count summary to %s.\n", param->primer_counts_file);
+        } else {
+            file_open = 1;
+        }
+
+        //print out the number of reads for each bed entry, bedgraph format
+        fprintf(bed_count_summary_fp, "#CHR\tLEFT\tRIGHT\tNAME\tSCORE\tSTRAND\tNUM_CLIPPED\n");
+        size_t refidx;
+        for (refidx = 0; refidx < num_bed_refs; refidx++) {
+            khiter_t itr = kh_get(bed_list_hash, bed_hash, bed_ref_list[refidx]);
+            if (itr >= kh_end(bed_hash)) {
+                fprintf(stderr, "[ampliconclip] error: %s has gone missing from the hash table\n", bed_ref_list[refidx]);
+                goto fail;
+            }
+            sites = &kh_val(bed_hash, itr);
+            int i;
+            for (i = 0; i < sites->length; i++) {
+                char* strand_out = param->use_strand ? (sites->bp[i].rev ? "-" : "+") : ".";
+                fprintf(bed_count_summary_fp, "%s\t%"PRId64"\t%"PRId64"\t%s\t%s\t%s\t%"PRId64"\n",
+                        kh_key(bed_hash, itr), sites->bp[i].left, sites->bp[i].right, sites->bp[i].name,
+                        sites->bp[i].score, strand_out, sites->bp[i].num_reads);
+            }
+        }
+        if (file_open) {
+            fclose(bed_count_summary_fp);
+        }
     }
 
     ret = 0;
 
 fail:
+    free(bed_ref_list);
     destroy_bed_hash(bed_hash);
     ks_free(&oat);
+    ks_free(&seq);
     sam_hdr_destroy(header);
     bam_destroy1(b);
     bam_destroy1(b_tmp);
@@ -920,24 +1049,26 @@ fail:
 static void usage(void) {
     fprintf(stderr, "Usage: samtools ampliconclip -b BED file <input.bam> -o <output.bam>\n\n");
     fprintf(stderr, "Option: \n");
-    fprintf(stderr, " -b  FILE            BED file of regions (eg amplicon primers) to be removed.\n");
-    fprintf(stderr, " -o  FILE            output file name (default stdout).\n");
-    fprintf(stderr, " -f  FILE            write stats to file name (default stderr)\n");
-    fprintf(stderr, " -u                  Output uncompressed data\n");
-    fprintf(stderr, " --soft-clip         soft clip amplicon primers from reads (default)\n");
-    fprintf(stderr, " --hard-clip         hard clip amplicon primers from reads.\n");
-    fprintf(stderr, " --both-ends         clip on both 5' and 3' ends.\n");
-    fprintf(stderr, " --strand            use strand data from BED file to match read direction.\n");
-    fprintf(stderr, " --clipped           only output clipped reads.\n");
-    fprintf(stderr, " --fail              mark unclipped, mapped reads as QCFAIL.\n");
-    fprintf(stderr, " --filter-len INT    do not output reads INT size or shorter.\n");
-    fprintf(stderr, " --fail-len   INT    mark as QCFAIL reads INT size or shorter.\n");
-    fprintf(stderr, " --no-excluded       do not write excluded reads (unmapped or QCFAIL).\n");
-    fprintf(stderr, " --rejects-file FILE file to write filtered reads.\n");
-    fprintf(stderr, " --original          for clipped entries add an OA tag with original data.\n");
-    fprintf(stderr, " --keep-tag          for clipped entries keep the old NM and MD tags.\n");
-    fprintf(stderr, " --tolerance         match region within this number of bases, default 5.\n");
-    fprintf(stderr, " --no-PG             do not add an @PG line.\n");
+    fprintf(stderr, " -b  FILE             BED file of regions (eg amplicon primers) to be removed.\n");
+    fprintf(stderr, " -o  FILE             output file name (default: stdout).\n");
+    fprintf(stderr, " -f  FILE             write stats to file name (default: stderr)\n");
+    fprintf(stderr, " -u                   Output uncompressed data\n");
+    fprintf(stderr, " --soft-clip          soft clip amplicon primers from reads (default)\n");
+    fprintf(stderr, " --hard-clip          hard clip amplicon primers from reads.\n");
+    fprintf(stderr, " --both-ends          clip on both 5' and 3' ends.\n");
+    fprintf(stderr, " --strand             use strand data from BED file to match read direction.\n");
+    fprintf(stderr, " --clipped            only output clipped reads.\n");
+    fprintf(stderr, " --fail               mark unclipped, mapped reads as QCFAIL.\n");
+    fprintf(stderr, " --filter-len INT     do not output reads INT size or shorter.\n");
+    fprintf(stderr, " --fail-len   INT     mark as QCFAIL reads INT size or shorter.\n");
+    fprintf(stderr, " --unmap-len  INT     unmap reads INT size or shorter, default 0.\n");
+    fprintf(stderr, " --no-excluded        do not write excluded reads (unmapped or QCFAIL).\n");
+    fprintf(stderr, " --rejects-file FILE  file to write filtered reads.\n");
+    fprintf(stderr, " --primer-counts FILE file to write read counts per bed entry (bedgraph format).\n");
+    fprintf(stderr, " --original           for clipped entries add an OA tag with original data.\n");
+    fprintf(stderr, " --keep-tag           for clipped entries keep the old NM and MD tags.\n");
+    fprintf(stderr, " --tolerance          match region within this number of bases, default 5.\n");
+    fprintf(stderr, " --no-PG              do not add an @PG line.\n");
     sam_global_opt_help(stderr, "-.O..@-.");
     fprintf(stderr, "\nAbout: Soft clips read alignments where they match BED file defined regions.\n"
                     "Default clipping is only on the 5' end.\n\n");
@@ -952,7 +1083,7 @@ int amplicon_clip_main(int argc, char **argv) {
     htsThreadPool p = {NULL, 0};
     samFile *in = NULL, *out = NULL, *reject = NULL;
     clipping_type clipping = soft_clip;
-    cl_param_t param = {1, 0, 0, 0, 0, -1, -1, 0, 0, 1, 5, NULL, NULL, NULL};
+    cl_param_t param = {1, 0, 0, 0, 0, -1, -1, 0, 0, 1, 5, 0, NULL, NULL, NULL};
 
     static const struct option lopts[] = {
         SAM_OPT_GLOBAL_OPTIONS('-', 0, 'O', 0, 0, '@'),
@@ -967,9 +1098,11 @@ int amplicon_clip_main(int argc, char **argv) {
         {"fail-len", required_argument, NULL, 1010},
         {"no-excluded", no_argument, NULL, 1011},
         {"rejects-file", required_argument, NULL, 1012},
-        {"original", no_argument, NULL, 1013},
-        {"keep-tag", no_argument, NULL, 1014},
-        {"tolerance", required_argument, NULL, 1015},
+        {"primer-counts", required_argument, NULL, 1013},
+        {"original", no_argument, NULL, 1014},
+        {"keep-tag", no_argument, NULL, 1015},
+        {"tolerance", required_argument, NULL, 1016},
+        {"unmap-len", required_argument, NULL, 1017},
         {NULL, 0, NULL, 0}
     };
 
@@ -990,9 +1123,11 @@ int amplicon_clip_main(int argc, char **argv) {
             case 1010: param.fail_len = atoi(optarg); break;
             case 1011: param.unmapped = 1; break;
             case 1012: param.rejects_file = optarg; break;
-            case 1013: param.oa_tag = 1; break;
-            case 1014: param.del_tag = 0; break;
-            case 1015: param.tol = atoi(optarg); break;
+            case 1013: param.primer_counts_file = optarg; break;
+            case 1014: param.oa_tag = 1; break;
+            case 1015: param.del_tag = 0; break;
+            case 1016: param.tol = atoi(optarg); break;
+            case 1017: param.unmap_len = atoi(optarg); break;
             default:  if (parse_sam_global_opt(c, optarg, lopts, &ga) == 0) break;
                       /* else fall-through */
             case '?': usage(); exit(1);
@@ -1011,7 +1146,7 @@ int amplicon_clip_main(int argc, char **argv) {
 
     if (param.tol < 0) {
         fprintf(stderr, "[ampliconclip] warning: invalid tolerance of %d,"
-                        " reseting tolerance to default of 5.\n", param.tol);
+                        " resetting tolerance to default of 5.\n", param.tol);
         param.tol = 5;
     }
 
