@@ -78,6 +78,7 @@ typedef struct {
     regex_t *bc_rgx;
     int read_groups;
     int json;
+    int dc;
 } md_param_t;
 
 typedef struct {
@@ -98,6 +99,7 @@ typedef struct read_queue_s {
     bam1_t *b;
     struct read_queue_s *duplicate;
     struct read_queue_s *original;
+    int dc;
     hts_pos_t pos;
     int dup_checked;
     int read_group;
@@ -122,6 +124,7 @@ typedef struct {
     int opt;
     int beg;
     int end;
+    int len;
 } check_t;
 
 typedef struct {
@@ -875,7 +878,10 @@ static int is_optical_duplicate(md_param_t *param, bam1_t *ori, bam1_t *dup, lon
         return ret;
     }
 
-    if (strncmp(original + o_beg, duplicate + d_beg, o_end - o_beg) == 0) {
+    int o_len = o_end - o_beg;
+    int d_len = d_end - d_beg;
+
+    if ((o_len == d_len) && memcmp(original + o_beg, duplicate + d_beg, o_len) == 0) {
         long xdiff, ydiff;
 
         if (ox > dx) {
@@ -918,7 +924,10 @@ static int optical_duplicate_partial(md_param_t *param, const char *name, const
         return ret;
     }
 
-    if (strncmp(name + o_beg, duplicate + d_beg, o_end - o_beg) == 0) {
+    int o_len = o_end - o_beg;
+    int d_len = d_end - d_beg;
+
+    if ((o_len == d_len) && memcmp(name + o_beg, duplicate + d_beg, o_len) == 0) {
         // the initial parts match, look at the numbers
         long xdiff, ydiff;
 
@@ -945,6 +954,7 @@ static int optical_duplicate_partial(md_param_t *param, const char *name, const
     c->y = dy;
     c->beg = d_beg;
     c->end = d_end;
+    c->len = d_end - d_beg;
 
     return ret;
 }
@@ -1156,9 +1166,15 @@ static int check_chain_against_original(md_param_t *param, khash_t(duplicates) *
 }
 
 
-static int xcoord_sort(const void *a, const void *b) {
+static int chain_sort(const void *a, const void *b) {
     check_t *ac = (check_t *) a;
     check_t *bc = (check_t *) b;
+    int ret;
+
+    if ((ret = ac->len - bc->len))
+        return ret;
+    else if ((ret = memcmp(bam_get_qname(ac->b) + ac->beg, bam_get_qname(bc->b) + bc->beg, ac->len)))
+        return ret;
 
     return (ac->x - bc->x);
 }
@@ -1170,106 +1186,113 @@ static int check_duplicate_chain(md_param_t *param, khash_t(duplicates) *dup_has
     int ret = 0;
     size_t curr = 0;
 
-    qsort(list->c, list->length, sizeof(list->c[0]), xcoord_sort);
+    qsort(list->c, list->length, sizeof(list->c[0]), chain_sort);
 
     while (curr < list->length - 1) {
-        check_t *current = &list->c[curr];
-        size_t count = curr;
-        char *cur_name = bam_get_qname(current->b);
-        int current_paired = (current->b->core.flag & BAM_FPAIRED) && !(current->b->core.flag & BAM_FMUNMAP);
+        check_t *base = &list->c[curr];
+        char *base_name = bam_get_qname(base->b);
+        int end_name_match = curr;
 
-        while (++count < list->length && (list->c[count].x - current->x <= param->opt_dist)) {
-            // while close enough along the x coordinate
-            check_t *chk = &list->c[count];
+        // find the end of the matching name parts
+        while (++end_name_match < list->length) {
+            check_t *chk = &list->c[end_name_match];
 
-            if (current->opt && chk->opt)
-                continue;
-
-            // if both are already optical duplicates there is no need to check again, otherwise...
-
-            long ydiff;
-
-            if (current->y > chk->y) {
-                ydiff = current->y - chk->y;
-            } else {
-                ydiff = chk->y - current->y;
-            }
-
-            if (ydiff > param->opt_dist)
-                continue;
-
-            // the number are right, check the names
-            if (strncmp(cur_name + current->beg, bam_get_qname(chk->b) + chk->beg, current->end - current->beg) != 0)
-                continue;
-
-            // optical duplicates
-            int chk_dup = 0;
-            int chk_paired = (chk->b->core.flag & BAM_FPAIRED) && !(chk->b->core.flag & BAM_FMUNMAP);
-
-            if (current_paired != chk_paired) {
-                if (!chk_paired) {
-                    // chk is single vs pair, this is a dup.
-                    chk_dup = 1;
-                }
-            } else {
-                // do it by scores
-                int64_t cur_score, chk_score;
-
-                if ((current->b->core.flag & BAM_FQCFAIL) != (chk->b->core.flag & BAM_FQCFAIL)) {
-                    if (current->b->core.flag & BAM_FQCFAIL) {
-                        cur_score = 0;
-                        chk_score = 1;
-                    } else {
-                        cur_score = 1;
-                        chk_score = 0;
-                    }
-                } else {
-                    cur_score = current->score;
-                    chk_score = chk->score;
-
-                    if (current_paired) {
-                        // they are pairs so add mate scores.
-                        chk_score += chk->mate_score;
-                        cur_score += current->mate_score;
-                    }
-                }
-
-                if (cur_score == chk_score) {
-                    if (strcmp(bam_get_qname(chk->b), cur_name) < 0) {
-                        chk_score++;
-                    } else {
-                        chk_score--;
-                    }
-                }
-
-                if (cur_score > chk_score) {
-                    chk_dup = 1;
-                }
-            }
-
-            if (chk_dup) {
-                // the duplicate is the optical duplicate
-                if (!chk->opt) { // only change if not already an optical duplicate
-                    if (optical_retag(param, dup_hash, chk->b, chk_paired, stats)) {
-                        ret = -1;
-                        goto fail;
-                    }
-
-                    chk->opt = 1;
-                }
-            } else {
-                if (!current->opt) {
-                    if (optical_retag(param, dup_hash, current->b, current_paired, stats)) {
-                        ret = -1;
-                        goto fail;
-                    }
-
-                    current->opt = 1;
-                }
-            }
+            if ((base->len == chk->len) && memcmp(base_name + base->beg, bam_get_qname(chk->b) + chk->beg, base->len) != 0)
+                break;
         }
 
-        curr++;
+        while (curr < end_name_match) {
+            size_t count = curr;
+            check_t *current = &list->c[curr];
+            int current_paired = (current->b->core.flag & BAM_FPAIRED) && !(current->b->core.flag & BAM_FMUNMAP);
+
+            while (++count < end_name_match && (list->c[count].x - current->x <= param->opt_dist)) {
+                // while close enough along the x coordinate
+                check_t *chk = &list->c[count];
+
+                if (current->opt && chk->opt)
+                    continue;
+
+                long ydiff;
+
+                if (current->y > chk->y) {
+                    ydiff = current->y - chk->y;
+                } else {
+                    ydiff = chk->y - current->y;
+                }
+
+                if (ydiff > param->opt_dist)
+                    continue;
+
+                // optical duplicates
+                int chk_dup = 0;
+                int chk_paired = (chk->b->core.flag & BAM_FPAIRED) && !(chk->b->core.flag & BAM_FMUNMAP);
+
+                if (current_paired != chk_paired) {
+                    if (!chk_paired) {
+                        // chk is single vs pair, this is a dup.
+                        chk_dup = 1;
+                    }
+                } else {
+                    // do it by scores
+                    int64_t cur_score, chk_score;
+
+                    if ((current->b->core.flag & BAM_FQCFAIL) != (chk->b->core.flag & BAM_FQCFAIL)) {
+                        if (current->b->core.flag & BAM_FQCFAIL) {
+                            cur_score = 0;
+                            chk_score = 1;
+                        } else {
+                            cur_score = 1;
+                            chk_score = 0;
+                        }
+                    } else {
+                        cur_score = current->score;
+                        chk_score = chk->score;
+
+                        if (current_paired) {
+                            // they are pairs so add mate scores.
+                            chk_score += chk->mate_score;
+                            cur_score += current->mate_score;
+                        }
+                    }
+
+                    if (cur_score == chk_score) {
+                        if (strcmp(bam_get_qname(chk->b), bam_get_qname(current->b)) < 0) {
+                            chk_score++;
+                        } else {
+                            chk_score--;
+                        }
+                    }
+
+                    if (cur_score > chk_score) {
+                        chk_dup = 1;
+                    }
+                }
+
+                if (chk_dup) {
+                    // the duplicate is the optical duplicate
+                    if (!chk->opt) { // only change if not already an optical duplicate
+                        if (optical_retag(param, dup_hash, chk->b, chk_paired, stats)) {
+                            ret = -1;
+                            goto fail;
+                        }
+
+                        chk->opt = 1;
+                    }
+                } else {
+                    if (!current->opt) {
+                        if (optical_retag(param, dup_hash, current->b, current_paired, stats)) {
+                            ret = -1;
+                            goto fail;
+                        }
+
+                        current->opt = 1;
+                    }
+                }
+            }
+
+            curr++;
+        }
     }
 
  fail:
@@ -1618,6 +1641,7 @@ static int bam_mark_duplicates(md_param_t *param) {
         in_read->original = NULL;
         in_read->dup_checked = 0;
         in_read->read_group = 0;
+        in_read->dc = 1;
 
         if (param->read_groups) {
             uint8_t *data;
@@ -1705,6 +1729,7 @@ static int bam_mark_duplicates(md_param_t *param) {
                         }
 
                         bp->p = in_read;
+                        bp->p->dc += 1;
 
                         if (mark_duplicates(param, dup_hash, bp->p->b, dup, in_read->read_group, &stats->single_optical, &opt_warnings))
                             goto fail;
@@ -1767,6 +1792,7 @@ static int bam_mark_duplicates(md_param_t *param) {
 
                     if (new_score + tie_add > old_score) { // swap reads
                         dup = bp->p->b;
+                        in_read->dc += bp->p->dc;
 
                         if (param->check_chain) {
 
@@ -1807,6 +1833,7 @@ static int bam_mark_duplicates(md_param_t *param) {
                         }
 
                         dup = in_read->b;
+                        bp->p->dc += 1;
                     }
 
                     if (mark_duplicates(param, dup_hash, bp->p->b, dup, in_read->read_group, &stats->optical, &opt_warnings))
@@ -1848,6 +1875,8 @@ static int bam_mark_duplicates(md_param_t *param) {
                             in_read->original = bp->p;
                         }
 
+                        bp->p->dc += 1;
+
                         if (mark_duplicates(param, dup_hash, bp->p->b, in_read->b, in_read->read_group, &stats->single_optical, &opt_warnings))
                             goto fail;
 
@@ -1862,6 +1891,7 @@ static int bam_mark_duplicates(md_param_t *param) {
                         // to the single hash and mark the other as duplicate
                         if (new_score > old_score) { // swap reads
                             dup = bp->p->b;
+                            in_read->dc += bp->p->dc;
 
                             if (param->check_chain) {
                                 in_read->duplicate = bp->p;
@@ -1879,6 +1909,7 @@ static int bam_mark_duplicates(md_param_t *param) {
                                 in_read->original = bp->p;
                             }
 
+                            bp->p->dc += 1;
                             dup = in_read->b;
                         }
 
@@ -1916,6 +1947,9 @@ static int bam_mark_duplicates(md_param_t *param) {
             }
 
             if (!param->remove_dups || !(in_read->b->core.flag & BAM_FDUP)) {
+                if (param->dc && !(in_read->b->core.flag & BAM_FDUP)) {
+                    bam_aux_update_int(in_read->b, "dc", in_read->dc);
+                }
                 if (param->supp) {
                     if (tmp_file_write(&temp, in_read->b)) {
                         print_error("markdup", "error, writing temp output failed.\n");
@@ -1979,12 +2013,20 @@ static int bam_mark_duplicates(md_param_t *param) {
             }
 
             if (!param->remove_dups || !(in_read->b->core.flag & BAM_FDUP)) {
+                if (param->dc && !(in_read->b->core.flag & BAM_FDUP)) {
+                    bam_aux_update_int(in_read->b, "dc",  in_read->dc);
+                }
+
                 if (param->supp) {
                     if (tmp_file_write(&temp, in_read->b)) {
                         print_error("markdup", "error, writing temp output failed on final write.\n");
                         goto fail;
                     }
                 } else {
+                    if (param->dc && !(in_read->b->core.flag & BAM_FDUP)) {
+                        bam_aux_update_int(in_read->b, "dc", in_read->dc);
+                    }
+
                     if (sam_write1(param->out, header, in_read->b) < 0) {
                         print_error("markdup", "error, writing output failed on final write.\n");
                         goto fail;
@@ -2046,6 +2088,10 @@ static int bam_mark_duplicates(md_param_t *param) {
             }
 
             if (!param->remove_dups || !(b->core.flag & BAM_FDUP)) {
+                if (param->dc && (b->core.flag & BAM_FDUP)) {
+                    uint8_t* data = bam_aux_get(b, "dc");
+                    if(data) bam_aux_del(b, data);
+                }
                 if (sam_write1(param->out, header, b) < 0) {
                     print_error("markdup", "error, writing final output failed.\n");
                     goto fail;
@@ -2181,6 +2227,7 @@ static int bam_mark_duplicates(md_param_t *param) {
     if (param->check_chain && (param->tag || param->opt_dist))
         free(dup_list.c);
 
+    free(idx_fn);
     free(stat_array);
     kh_destroy(reads, pair_hash);
     kh_destroy(reads, single_hash);
@@ -2207,6 +2254,7 @@ static int bam_mark_duplicates(md_param_t *param) {
     if (param->check_chain && (param->tag || param->opt_dist))
         free(dup_list.c);
 
+    free(idx_fn);
     free(stat_array);
     kh_destroy(reads, pair_hash);
     kh_destroy(reads, single_hash);
@@ -2244,6 +2292,7 @@ static int markdup_usage(void) {
     fprintf(samtools_stderr, "  --use-read-groups  Use the read group tags in duplicate matching.\n");
     fprintf(samtools_stderr, "  -t                 Mark primary duplicates with the name of the original in a \'do\' tag."
                                         " Mainly for information and debugging.\n");
+    fprintf(samtools_stderr, "  --duplicate-count  Record the original primary read duplication count(include itself) in a \'dc\' tag.\n");
 
     sam_global_opt_help(samtools_stderr, "-.O..@..");
 
@@ -2265,7 +2314,7 @@ int bam_markdup(int argc, char **argv) {
     char *regex = NULL, *bc_regex = NULL;
     char *regex_order = "txy";
     md_param_t param = {NULL, NULL, NULL, 0, 300, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-                        1, NULL, NULL, NULL, NULL, 0, 0, 0, NULL, NULL, 0, 0};
+                        1, NULL, NULL, NULL, NULL, 0, 0, 0, NULL, NULL, 0, 0, 0};
 
     static const struct option lopts[] = {
         SAM_OPT_GLOBAL_OPTIONS('-', 0, 'O', 0, 0, '@'),
@@ -2280,6 +2329,7 @@ int bam_markdup(int argc, char **argv) {
         {"barcode-rgx", required_argument, NULL, 1008},
         {"use-read-groups", no_argument, NULL, 1009},
         {"json", no_argument, NULL, 1010},
+        {"duplicate-count", no_argument, NULL, 1011},
         {NULL, 0, NULL, 0}
     };
 
@@ -2316,6 +2366,7 @@ int bam_markdup(int argc, char **argv) {
             case 1008: bc_name = 1, bc_regex = optarg; break;
             case 1009: param.read_groups = 1; break;
             case 1010: param.json = 1; param.do_stats = 1; break;
+            case 1011: param.dc = 1; break;
             default: if (parse_sam_global_opt(c, optarg, lopts, &ga) == 0) break;
             /* else fall-through */
             case '?': return markdup_usage();
