@@ -129,6 +129,11 @@ int cram_reheader(cram_fd *in, sam_hdr_t *h, const char *arg_list, int no_pg)
     if (!h)
         return ret;
 
+    // Match output version number with input file.
+    char vers[99];
+    sprintf(vers, "%d.%d", cram_major_vers(in), cram_minor_vers(in));
+    cram_set_option(out, CRAM_OPT_VERSION, vers);
+
     // Attempt to fill out a cram->refs[] array from @SQ headers
     sam_hdr_t *cram_h = sam_hdr_dup(h);
     if (!cram_h)
@@ -382,7 +387,7 @@ int cram_reheader_inplace3(cram_fd *fd, sam_hdr_t *h, const char *arg_list,
     cram_block_append(b, (void *)sam_hdr_str(cram_h), header_len);
     cram_block_update_size(b);
 
-    cram_compress_block(fd, b, NULL, -1, -1);
+    cram_compress_block(fd, b, NULL, -1, 9);
 
     if (hseek(cram_fd_get_fp(fd), 26, SEEK_SET) != 26)
         goto err;
@@ -533,7 +538,7 @@ cleanup:
     return h;
 }
 
-int samtools_main_reheader(int argc, char *argv[])
+int main_reheader(int argc, char *argv[])
 {
     int inplace = 0, r, no_pg = 0, c, skip_header = 0;
     sam_hdr_t *h;
