@@ -31,16 +31,20 @@ DEALINGS IN THE SOFTWARE.  */
 #include <unistd.h>
 #include <zlib.h>
 #include <getopt.h>
+#include "htslib/khash.h"
 #include "htslib/kseq.h"
 #include "htslib/hts.h"
+#include "samtools.h"
 
+KHASH_SET_INIT_STR(str)
 KSEQ_INIT(gzFile, gzread)
 
 typedef struct _args_t
 {
-    char *output_fname, *fname;
+    char *output_fname, *alt_fname;
     char *assembly, *species, *uri;
     int  alias, header;
+    khash_t(str) *is_alt;
 }
 args_t;
 
@@ -55,14 +59,14 @@ static void write_dict(const char *fn, args_t *args)
 
     fp = strcmp(fn, "-") ? gzopen(fn, "r") : gzdopen(fileno(stdin), "r");
     if (fp == 0) {
-        fprintf(samtools_stderr, "dict: %s: No such file or directory\n", fn);
+        print_error_errno("dict", "Cannot open %s", fn);
         samtools_exit(1);
     }
     FILE *out = samtools_stdout;
     if (args->output_fname) {
         out = fopen(args->output_fname, "w");
         if (out == NULL) {
-          fprintf(samtools_stderr, "dict: %s: Cannot open file for writing\n", args->output_fname);
+          print_error_errno("dict", "Cannot open %s for writing", args->output_fname);
           samtools_exit(1);
         }
     }
@@ -82,6 +86,8 @@ static void write_dict(const char *fn, args_t *args)
         hts_md5_final(digest, md5);
         hts_md5_hex(hex, digest);
         fprintf(out, "@SQ\tSN:%s\tLN:%d\tM5:%s", seq->name.s, k, hex);
+        if (args->is_alt && kh_get(str, args->is_alt, seq->name.s) != kh_end(args->is_alt))
+            fprintf(out, "\tAH:*");
         if (args->alias) {
             const char *name = seq->name.s;
             if (strncmp(name, "chr", 3) == 0) {
@@ -118,6 +124,34 @@ static void write_dict(const char *fn, args_t *args)
     gzclose(fp);
 }
 
+static void read_alt_file(khash_t(str) *is_alt, const char *fname)
+{
+    htsFile *fp = hts_open(fname, "r");
+    if (fp == NULL) {
+        print_error_errno("dict", "Cannot open %s", fname);
+        samtools_exit(1);
+    }
+
+    // .alt files are in a SAM-like format, but we don't use sam_read1()
+    // as these files may not have a complete set of @SQ headers.
+
+    kstring_t str = KS_INITIALIZE;
+    while (hts_getline(fp, KS_SEP_LINE, &str) >= 0) {
+        if (str.l == 0 || str.s[0] == '@') continue;
+
+        char *tab = strchr(str.s, '\t');
+        if (tab) *tab = '\0';
+
+        int ret;
+        char *seqname = strdup(str.s);
+        kh_put(str, is_alt, seqname, &ret);
+        if (ret == 0) free(seqname); // Already present
+    }
+
+    ks_free(&str);
+    hts_close(fp);
+}
+
 static int dict_usage(void)
 {
     fprintf(samtools_stderr, "\n");
@@ -127,6 +161,7 @@ static int dict_usage(void)
     fprintf(samtools_stderr, "         -A, --alias, --alternative-name\n");
     fprintf(samtools_stderr, "                               add AN tag by adding/removing 'chr'\n");
     fprintf(samtools_stderr, "         -H, --no-header       do not print @HD line\n");
+    fprintf(samtools_stderr, "         -l, --alt FILE        add AH:* tag to alternate locus sequences\n");
     fprintf(samtools_stderr, "         -o, --output FILE     file to write out dict file [samtools_stdout]\n");
     fprintf(samtools_stderr, "         -s, --species STR     species\n");
     fprintf(samtools_stderr, "         -u, --uri STR         URI [file:///abs/path/to/file.fa]\n");
@@ -144,6 +179,7 @@ int dict_main(int argc, char *argv[])
         {"help", no_argument, NULL, 'h'},
         {"no-header", no_argument, NULL, 'H'},
         {"alias", no_argument, NULL, 'A'},
+        {"alt", required_argument, NULL, 'l'},
         {"alternative-name", no_argument, NULL, 'A'},
         {"assembly", required_argument, NULL, 'a'},
         {"species", required_argument, NULL, 's'},
@@ -152,12 +188,13 @@ int dict_main(int argc, char *argv[])
         {NULL, 0, NULL, 0}
     };
     int c;
-    while ( (c=getopt_long(argc,argv,"?AhHa:s:u:o:",loptions,NULL))>0 )
+    while ( (c=getopt_long(argc,argv,"?AhHa:l:s:u:o:",loptions,NULL))>0 )
     {
         switch (c)
         {
             case 'A': args->alias = 1; break;
             case 'a': args->assembly = optarg; break;
+            case 'l': args->alt_fname = optarg; break;
             case 's': args->species = optarg; break;
             case 'u': args->uri = optarg; break;
             case 'o': args->output_fname = optarg; break;
@@ -175,7 +212,20 @@ int dict_main(int argc, char *argv[])
     }
     else fname = argv[optind];
 
+    if (args->alt_fname) {
+        args->is_alt = kh_init(str);
+        read_alt_file(args->is_alt, args->alt_fname);
+    }
+
     write_dict(fname, args);
+
+    if (args->is_alt) {
+        khint_t k;
+        for (k = 0; k < kh_end(args->is_alt); ++k)
+            if (kh_exist(args->is_alt, k)) free((char *) kh_key(args->is_alt, k));
+        kh_destroy(str, args->is_alt);
+    }
+
     free(args);
     return 0;
 }
