@@ -3,7 +3,7 @@
 /*  padding.c -- depad subcommand.
 
     Copyright (C) 2011, 2012 Broad Institute.
-    Copyright (C) 2014-2016 Genome Research Ltd.
+    Copyright (C) 2014-2016, 2019-2020 Genome Research Ltd.
     Portions copyright (C) 2012, 2013 Peter Cock, The James Hutton Institute.
 
     Author: Heng Li <lh3@sanger.ac.uk>
@@ -31,39 +31,57 @@ DEALINGS IN THE SOFTWARE.  */
 #include <string.h>
 #include <assert.h>
 #include <unistd.h>
+#include <inttypes.h>
 #include <htslib/kstring.h>
 #include <htslib/sam.h>
 #include <htslib/faidx.h>
-#include "sam_header.h"
 #include "sam_opts.h"
 #include "samtools.h"
 
 #define bam_reg2bin(b,e) hts_reg2bin((b),(e), 14, 5)
 
-// The one and only function needed from sam.c.
-// Explicitly here to avoid including bam.h translation layer.
-extern char *samfaipath(const char *fn_ref);
-
-static void replace_cigar(bam1_t *b, int n, uint32_t *cigar)
+static int replace_cigar(bam1_t *b, uint32_t n, uint32_t *cigar)
 {
+    int diff = 0;
     if (n != b->core.n_cigar) {
         int o = b->core.l_qname + b->core.n_cigar * 4;
-        if (b->l_data + (n - b->core.n_cigar) * 4 > b->m_data) {
-            b->m_data = b->l_data + (n - b->core.n_cigar) * 4;
-            kroundup32(b->m_data);
-            b->data = (uint8_t*)realloc(b->data, b->m_data);
+        if (n > b->core.n_cigar) {
+            diff = (n - b->core.n_cigar) * 4;
+            if ((INT_MAX - b->l_data)/4 < (n - b->core.n_cigar)) {
+                fprintf(samtools_stderr, "[depad] ERROR: BAM record too big\n");
+                return -1;
+            }
+            if (b->l_data + diff > b->m_data) {
+                b->m_data = b->l_data + diff;
+                kroundup32(b->m_data);
+                uint8_t *tmp = (uint8_t*)realloc(b->data, b->m_data);
+                if (!tmp) {
+                    fprintf(samtools_stderr, "[depad] ERROR: Memory allocation failure.\n");
+                    return -1;
+                }
+                b->data = tmp;
+            }
+        } else {
+            diff = -(int)((b->core.n_cigar - n) * 4);
         }
         memmove(b->data + b->core.l_qname + n * 4, b->data + o, b->l_data - o);
-        memcpy(b->data + b->core.l_qname, cigar, n * 4);
-        b->l_data += (n - b->core.n_cigar) * 4;
         b->core.n_cigar = n;
-    } else memcpy(b->data + b->core.l_qname, cigar, n * 4);
+    }
+
+    memcpy(b->data + b->core.l_qname, cigar, n * 4);
+    b->l_data += diff;
+
+    return 0;
 }
 
 #define write_cigar(_c, _n, _m, _v) do { \
         if (_n == _m) { \
             _m = _m? _m<<1 : 4; \
             _c = (uint32_t*)realloc(_c, _m * 4); \
+            if (!(_c)) { \
+                fprintf(samtools_stderr, "[depad] ERROR: Memory allocation failure.\n"); \
+                return -1; \
+            } \
         } \
         _c[_n++] = (_v); \
     } while (0)
@@ -109,15 +127,15 @@ static int unpad_seq(bam1_t *b, kstring_t *s)
     return length != s->l;
 }
 
-int load_unpadded_ref(faidx_t *fai, char *ref_name, int ref_len, kstring_t *seq)
+int load_unpadded_ref(faidx_t *fai, const char *ref_name, hts_pos_t ref_len, kstring_t *seq)
 {
     char base;
     char *fai_ref = 0;
-    int fai_ref_len = 0, k;
+    hts_pos_t fai_ref_len = 0, k;
 
-    fai_ref = fai_fetch(fai, ref_name, &fai_ref_len);
+    fai_ref = fai_fetch64(fai, ref_name, &fai_ref_len);
     if (fai_ref_len != ref_len) {
-        fprintf(samtools_stderr, "[depad] ERROR: FASTA sequence %s length %i, expected %i\n", ref_name, fai_ref_len, ref_len);
+        fprintf(samtools_stderr, "[depad] ERROR: FASTA sequence %s length %"PRIhts_pos", expected %"PRIhts_pos"\n", ref_name, fai_ref_len, ref_len);
         free(fai_ref);
         return -1;
     }
@@ -143,16 +161,16 @@ int load_unpadded_ref(faidx_t *fai, char *ref_name, int ref_len, kstring_t *seq)
     return 0;
 }
 
-int get_unpadded_len(faidx_t *fai, char *ref_name, int padded_len)
+hts_pos_t get_unpadded_len(faidx_t *fai, const char *ref_name, hts_pos_t padded_len)
 {
     char base;
     char *fai_ref = 0;
-    int fai_ref_len = 0, k;
-    int bases=0, gaps=0;
+    hts_pos_t fai_ref_len = 0, k;
+    hts_pos_t bases=0, gaps=0;
 
-    fai_ref = fai_fetch(fai, ref_name, &fai_ref_len);
+    fai_ref = fai_fetch64(fai, ref_name, &fai_ref_len);
     if (fai_ref_len != padded_len) {
-        fprintf(samtools_stderr, "[depad] ERROR: FASTA sequence '%s' length %i, expected %i\n", ref_name, fai_ref_len, padded_len);
+        fprintf(samtools_stderr, "[depad] ERROR: FASTA sequence '%s' length %"PRIhts_pos", expected %"PRIhts_pos"\n", ref_name, fai_ref_len, padded_len);
         free(fai_ref);
         return -1;
     }
@@ -187,13 +205,14 @@ static inline int * update_posmap(int *posmap, kstring_t ref)
     return posmap;
 }
 
-int bam_pad2unpad(samFile *in, samFile *out,  bam_hdr_t *h, faidx_t *fai)
+int bam_pad2unpad(samFile *in, samFile *out,  sam_hdr_t *h, faidx_t *fai)
 {
     bam1_t *b = 0;
     kstring_t r, q;
     int r_tid = -1;
     uint32_t *cigar2 = 0;
-    int ret = 0, n2 = 0, m2 = 0, *posmap = 0;
+    int ret = 0, *posmap = 0;
+    uint32_t n2 = 0, m2 = 0;
 
     b = bam_init1();
     if (!b) {
@@ -209,21 +228,21 @@ int bam_pad2unpad(samFile *in, samFile *out,  bam_hdr_t *h, faidx_t *fai)
 
         uint32_t *cigar = bam_get_cigar(b);
         n2 = 0;
-        if (b->core.pos == 0 && b->core.tid >= 0 && strcmp(bam_get_qname(b), h->target_name[b->core.tid]) == 0) {
+        if (b->core.pos == 0 && b->core.tid >= 0 && strcmp(bam_get_qname(b), sam_hdr_tid2name(h, b->core.tid)) == 0) {
             // fprintf(samtools_stderr, "[depad] Found embedded reference '%s'\n", bam_get_qname(b));
             r_tid = b->core.tid;
             if (0!=unpad_seq(b, &r)) {
                 fprintf(samtools_stderr, "[depad] ERROR: Problem parsing SEQ and/or CIGAR in reference %s\n", bam_get_qname(b));
                 return -1;
             };
-            if (h->target_len[r_tid] != r.l) {
-                fprintf(samtools_stderr, "[depad] ERROR: (Padded) length of '%s' is %u in BAM header, but %llu in embedded reference\n", bam_get_qname(b), h->target_len[r_tid], (unsigned long long)(r.l));
+            if (sam_hdr_tid2len(h, r_tid) != r.l) {
+                fprintf(samtools_stderr, "[depad] ERROR: (Padded) length of '%s' is %"PRId64" in BAM header, but %zu in embedded reference\n", bam_get_qname(b), (int64_t) sam_hdr_tid2len(h, r_tid), r.l);
                 return -1;
             }
             if (fai) {
                 // Check the embedded reference matches the FASTA file
-                if (load_unpadded_ref(fai, h->target_name[b->core.tid], h->target_len[b->core.tid], &q)) {
-                    fprintf(samtools_stderr, "[depad] ERROR: Failed to load embedded reference '%s' from FASTA\n", h->target_name[b->core.tid]);
+                if (load_unpadded_ref(fai, sam_hdr_tid2name(h, b->core.tid), sam_hdr_tid2len(h, b->core.tid), &q)) {
+                    fprintf(samtools_stderr, "[depad] ERROR: Failed to load embedded reference '%s' from FASTA\n", sam_hdr_tid2name(h, b->core.tid));
                     return -1;
                 }
                 assert(r.l == q.l);
@@ -232,7 +251,7 @@ int bam_pad2unpad(samFile *in, samFile *out,  bam_hdr_t *h, faidx_t *fai)
                     if (r.s[i] != q.s[i]) {
                         // Show gaps as ASCII 45
                         fprintf(samtools_stderr, "[depad] ERROR: Embedded sequence and reference FASTA don't match for %s base %i, '%c' vs '%c'\n",
-                            h->target_name[b->core.tid], i+1,
+                            sam_hdr_tid2name(h, b->core.tid), i+1,
                             r.s[i] ? seq_nt16_str[(int)r.s[i]] : 45,
                             q.s[i] ? seq_nt16_str[(int)q.s[i]] : 45);
                         return -1;
@@ -240,7 +259,8 @@ int bam_pad2unpad(samFile *in, samFile *out,  bam_hdr_t *h, faidx_t *fai)
                 }
             }
             write_cigar(cigar2, n2, m2, bam_cigar_gen(b->core.l_qseq, BAM_CMATCH));
-            replace_cigar(b, n2, cigar2);
+            if (replace_cigar(b, n2, cigar2) < 0)
+                return -1;
             posmap = update_posmap(posmap, r);
         } else if (b->core.n_cigar > 0) {
             int i, k, op;
@@ -251,15 +271,15 @@ int bam_pad2unpad(samFile *in, samFile *out,  bam_hdr_t *h, faidx_t *fai)
                 ; // good case, reference available
                 //fprintf(samtools_stderr, "[depad] Have ref '%s' for read '%s'\n", h->target_name[b->core.tid], bam_get_qname(b));
             } else if (fai) {
-                if (load_unpadded_ref(fai, h->target_name[b->core.tid], h->target_len[b->core.tid], &r)) {
-                    fprintf(samtools_stderr, "[depad] ERROR: Failed to load '%s' from reference FASTA\n", h->target_name[b->core.tid]);
+                if (load_unpadded_ref(fai, sam_hdr_tid2name(h, b->core.tid), sam_hdr_tid2len(h, b->core.tid), &r)) {
+                    fprintf(samtools_stderr, "[depad] ERROR: Failed to load '%s' from reference FASTA\n", sam_hdr_tid2name(h, b->core.tid));
                     return -1;
                 }
                 posmap = update_posmap(posmap, r);
                 r_tid = b->core.tid;
                 // fprintf(samtools_stderr, "[depad] Loaded %s from FASTA file\n", h->target_name[b->core.tid]);
             } else {
-                fprintf(samtools_stderr, "[depad] ERROR: Missing %s embedded reference sequence (and no FASTA file)\n", h->target_name[b->core.tid]);
+                fprintf(samtools_stderr, "[depad] ERROR: Missing %s embedded reference sequence (and no FASTA file)\n", sam_hdr_tid2name(h, b->core.tid));
                 return -1;
             }
             if (0!=unpad_seq(b, &q)) {
@@ -326,7 +346,8 @@ int bam_pad2unpad(samFile *in, samFile *out,  bam_hdr_t *h, faidx_t *fai)
             for (i = k = 0; i < n2; ++i)
                 if (cigar2[i]) cigar2[k++] = cigar2[i];
             n2 = k;
-            replace_cigar(b, n2, cigar2);
+            if (replace_cigar(b, n2, cigar2) < 0)
+                return -1;
         }
         /* Even unmapped reads can have a POS value, e.g. if their mate was mapped */
         if (b->core.pos != -1) b->core.pos = posmap[b->core.pos];
@@ -345,19 +366,19 @@ int bam_pad2unpad(samFile *in, samFile *out,  bam_hdr_t *h, faidx_t *fai)
             /* Nasty case, Must load alternative posmap */
             // fprintf(samtools_stderr, "[depad] Loading reference '%s' temporarily\n", h->target_name[b->core.mtid]);
             if (!fai) {
-                fprintf(samtools_stderr, "[depad] ERROR: Needed reference %s sequence for mate (and no FASTA file)\n", h->target_name[b->core.mtid]);
+                fprintf(samtools_stderr, "[depad] ERROR: Needed reference %s sequence for mate (and no FASTA file)\n", sam_hdr_tid2name(h, b->core.mtid));
                 return -1;
             }
             /* Temporarily load the other reference sequence */
-            if (load_unpadded_ref(fai, h->target_name[b->core.mtid], h->target_len[b->core.mtid], &r)) {
-                fprintf(samtools_stderr, "[depad] ERROR: Failed to load '%s' from reference FASTA\n", h->target_name[b->core.mtid]);
+            if (load_unpadded_ref(fai, sam_hdr_tid2name(h, b->core.mtid), sam_hdr_tid2len(h, b->core.mtid), &r)) {
+                fprintf(samtools_stderr, "[depad] ERROR: Failed to load '%s' from reference FASTA\n", sam_hdr_tid2name(h, b->core.mtid));
                 return -1;
             }
             posmap = update_posmap(posmap, r);
             b->core.mpos = posmap[b->core.mpos];
             /* Restore the reference and posmap*/
-            if (load_unpadded_ref(fai, h->target_name[b->core.tid], h->target_len[b->core.tid], &r)) {
-                fprintf(samtools_stderr, "[depad] ERROR: Failed to load '%s' from reference FASTA\n", h->target_name[b->core.tid]);
+            if (load_unpadded_ref(fai, sam_hdr_tid2name(h, b->core.tid), sam_hdr_tid2len(h, b->core.tid), &r)) {
+                fprintf(samtools_stderr, "[depad] ERROR: Failed to load '%s' from reference FASTA\n", sam_hdr_tid2name(h, b->core.tid));
                 return -1;
             }
             posmap = update_posmap(posmap, r);
@@ -376,126 +397,47 @@ int bam_pad2unpad(samFile *in, samFile *out,  bam_hdr_t *h, faidx_t *fai)
         ret = 1;
     }
     free(r.s); free(q.s); free(posmap);
+    free(cigar2);
     bam_destroy1(b);
     return ret;
 }
 
-bam_hdr_t * fix_header(bam_hdr_t *old, faidx_t *fai)
+sam_hdr_t * fix_header(sam_hdr_t *old, faidx_t *fai)
 {
-    int i = 0, unpadded_len = 0;
-    bam_hdr_t *header = 0 ;
-    unsigned short ln_found;
+    int i = 0, ret = 0;
+    hts_pos_t unpadded_len = 0;
+    sam_hdr_t *header = sam_hdr_dup(old);
+    if (!header)
+        return NULL;
 
-    header = bam_hdr_dup(old);
-    for (i = 0; i < old->n_targets; ++i) {
-        unpadded_len = get_unpadded_len(fai, old->target_name[i], old->target_len[i]);
+    int nref = sam_hdr_nref(old);
+    char len_buf[64];
+
+    for (i = 0; i < nref; ++i) {
+        unpadded_len = get_unpadded_len(fai, sam_hdr_tid2name(old, i), sam_hdr_tid2len(old, i));
         if (unpadded_len < 0) {
-            fprintf(samtools_stderr, "[depad] ERROR getting unpadded length of '%s', padded length %i\n", old->target_name[i], old->target_len[i]);
+            fprintf(samtools_stderr, "[depad] ERROR getting unpadded length of '%s', padded length %"PRIhts_pos"\n", sam_hdr_tid2name(old, i), (hts_pos_t) sam_hdr_tid2len(old, i));
+        } else if (unpadded_len > sam_hdr_tid2len(old, i)) {
+            fprintf(samtools_stderr, "[depad] New unpadded length of '%s' is larger than the padded length (%"PRIhts_pos" > %"PRIhts_pos")\n",
+                    sam_hdr_tid2name(old, i), unpadded_len,
+                    (hts_pos_t) sam_hdr_tid2len(old, i));
+            ret = 1;
         } else {
-            header->target_len[i] = unpadded_len;
+            sprintf(len_buf, "%"PRIhts_pos"", unpadded_len);
+            if ((ret |= sam_hdr_update_line(header, "SQ", "SN", sam_hdr_tid2name(header, i), "LN", len_buf, NULL)))
+                fprintf(samtools_stderr, "[depad] Error updating length of '%s' from %"PRIhts_pos" to %"PRIhts_pos"\n",
+                        sam_hdr_tid2name(header, i),
+                        (hts_pos_t) sam_hdr_tid2len(header, i),
+                        unpadded_len);
             //fprintf(samtools_stderr, "[depad] Recalculating '%s' length %i -> %i\n", old->target_name[i], old->target_len[i], header->target_len[i]);
         }
     }
-    /* Duplicating the header allocated new buffer for header string */
-    /* After modifying the @SQ lines it will only get smaller, since */
-    /* the LN entries will be the same or shorter, and we'll remove */
-    /* any MD entries (MD5 checksums). */
-    assert(strlen(old->text) == strlen(header->text));
-    assert (0==strcmp(old->text, header->text));
-    const char *text;
-    text = old->text;
-    header->text[0] = '\0'; /* Resuse the allocated buffer */
-    char * newtext = header->text;
-    char * end=NULL;
-    while (text[0]=='@') {
-        end = strchr(text, '\n');
-        assert(end != 0);
-        if (text[1]=='S' && text[2]=='Q' && text[3]=='\t') {
-            const char *cp = text+3;
-            char *name = strstr(text, "\tSN:");
-            char *name_end;
-            if (!name) {
-                fprintf(samtools_stderr, "Unable to find SN: header field\n");
-                return NULL;
-            }
-            name += 4;
-            for (name_end = name; name_end != end && *name_end != '\t'; name_end++);
-            strcat(newtext, "@SQ");
-            ln_found = 0;
 
-            /* Parse the @SQ lines */
-            while (cp != end) {
-                if (!ln_found && end-cp >= 2 && strncmp(cp, "LN", 2) == 0) {
-                    // Rewrite the length
-                    char len_buf[100];
-                    int tid;
-                    unsigned int old_length, new_length;
-                    const char *old_cp = cp;
-
-                    ln_found = 1;
-
-                    while (cp != end && *cp++ != '\t');
-                    old_length = (int)(cp - old_cp);
-
-                    for (tid = 0; tid < header->n_targets; tid++) {
-                        // may want to hash this, but new header API incoming.
-                        if (strncmp(name, header->target_name[tid], name_end - name) == 0) {
-                            new_length = sprintf(len_buf, "LN:%d", header->target_len[tid]);
-                            if (new_length <= old_length) {
-                                strcat(newtext, len_buf);
-                            }
-                            else {
-                                fprintf(samtools_stderr, "LN value of the reference is larger than the original!\n");
-                                exit(1);
-                            }
-                            break;
-                        }
-                    }
-
-                    if (cp != end)
-                        strcat(newtext, "\t");
-                } else if (end-cp >= 2 &&
-                           ((ln_found && strncmp(cp, "LN", 2) == 0) ||
-                            strncmp(cp, "M5", 2) == 0 ||
-                            strncmp(cp, "UR", 2) == 0))
-                {
-                    // skip secondary LNs
-                    // MD5 changed during depadding; ditch it.
-                    // URLs are also invalid.
-                    while (cp != end && *cp++ != '\t');
-                } else {
-                    // Otherwise copy this sub-field verbatim
-                    const char *cp_start = cp;
-                    while (cp != end && *cp++ != '\t');
-                    strncat(newtext, cp_start, cp-cp_start);
-                }
-            }
-
-            // Add newline, replacing trailing '\t' if last on line was the LN:
-            char *text_end = newtext + strlen(newtext);
-            if (text_end[-1] == '\t')
-                text_end[-1] = '\n';
-            else
-                *text_end++ = '\n', *text_end = '\0';
-        } else {
-            /* Copy this line to the new header */
-            strncat(newtext, text, end - text + 1);
-        }
-        text = end + 1;
+    if (ret) {
+        sam_hdr_destroy(header);
+        return NULL;
     }
-    assert (text[0]=='\0');
-    /* Check we didn't overflow the buffer */
-    assert (strlen(header->text) <= strlen(old->text));
-    if (strlen(header->text) < header->l_text) {
-        //fprintf(samtools_stderr, "[depad] Reallocating header buffer\n");
-        assert (newtext == header->text);
-        newtext = malloc(strlen(header->text) + 1);
-        strcpy(newtext, header->text);
-        free(header->text);
-        header->text = newtext;
-        header->l_text = strlen(newtext);
-    }
-    //fprintf(samtools_stderr, "[depad] Here is the new header (pending @SQ lines),\n\n%s\n(end)\n", header->text);
+
     return header;
 }
 
@@ -504,15 +446,17 @@ static int usage(int is_long_help);
 int main_pad2unpad(int argc, char *argv[])
 {
     samFile *in = 0, *out = 0;
-    bam_hdr_t *h = 0, *h_fix = 0;
+    sam_hdr_t *h = 0, *h_fix = 0;
     faidx_t *fai = 0;
-    int c, compress_level = -1, is_long_help = 0;
-    char in_mode[5], out_mode[6], *fn_out = 0, *fn_list = 0;
+    int c, compress_level = -1, is_long_help = 0, no_pg = 0;
+    char in_mode[5], out_mode[6], *fn_out = 0, *fn_fai = 0, *fn_out_idx = NULL;
     int ret=0;
+    char *arg_list = NULL;
     sam_global_args ga = SAM_GLOBAL_ARGS_INIT;
 
     static const struct option lopts[] = {
         SAM_OPT_GLOBAL_OPTIONS('-', 0, 0, 0, 'T', '-'),
+        {"no-PG", no_argument, NULL, 1},
         { NULL, 0, NULL, 0 }
     };
 
@@ -534,6 +478,7 @@ int main_pad2unpad(int argc, char *argv[])
             if (ga.out.format == unknown_format)
                 hts_parse_format(&ga.out, "bam");
             break;
+        case 1: no_pg = 1; break;
         case '?': is_long_help = 1; break;
         default:  if (parse_sam_global_opt(c, optarg, lopts, &ga) == 0) break;
             fprintf(samtools_stderr, "[bam_fillmd] unrecognized option '-%c'\n\n", c);
@@ -551,8 +496,8 @@ int main_pad2unpad(int argc, char *argv[])
 
     // Load FASTA reference (also needed for SAM -> BAM if missing header)
     if (ga.reference) {
-        fn_list = samfaipath(ga.reference);
-        fai = fai_load(ga.reference);
+        fn_fai = fai_path(ga.reference);
+        fai = fai_load3(ga.reference, fn_fai, NULL, FAI_CREATE);
     }
     // open file handlers
     if ((in = sam_open_format(argv[optind], in_mode, &ga.in)) == 0) {
@@ -560,8 +505,8 @@ int main_pad2unpad(int argc, char *argv[])
         ret = 1;
         goto depad_end;
     }
-    if (fn_list && hts_set_fai_filename(in, fn_list) != 0) {
-        fprintf(samtools_stderr, "[depad] failed to load reference file \"%s\".\n", fn_list);
+    if (fn_fai && hts_set_fai_filename(in, fn_fai) != 0) {
+        fprintf(samtools_stderr, "[depad] failed to load reference file \"%s\".\n", fn_fai);
         ret = 1;
         goto depad_end;
     }
@@ -571,7 +516,11 @@ int main_pad2unpad(int argc, char *argv[])
         goto depad_end;
     }
     if (fai) {
-        h_fix = fix_header(h, fai);
+        if (!(h_fix = fix_header(h, fai))){
+            fprintf(samtools_stderr, "[depad] failed to fix the header from\n");
+            ret = 1;
+            goto depad_end;
+        }
     } else {
         fprintf(samtools_stderr, "[depad] Warning - reference lengths will not be corrected without FASTA reference\n");
         h_fix = h;
@@ -589,25 +538,61 @@ int main_pad2unpad(int argc, char *argv[])
     if (ga.out.format == cram)
         hts_set_opt(out, CRAM_OPT_NO_REF, 1);
 
+    if (!no_pg) {
+        if(!(arg_list = stringify_argv(argc+1, argv-1))) {
+            fprintf(samtools_stderr, "[depad] failed to create arg_list\n");
+            ret = 1;
+            goto depad_end;
+            }
+
+        if (sam_hdr_add_pg(h_fix, "samtools",
+                           "VN", samtools_version(),
+                           arg_list ? "CL": NULL,
+                           arg_list ? arg_list : NULL,
+                           NULL)) {
+            fprintf(samtools_stderr, "[depad] failed to add PG line to header\n");
+            ret = 1;
+            goto depad_end;
+        }
+    }
+
     if (sam_hdr_write(out, h_fix) != 0) {
         fprintf(samtools_stderr, "[depad] failed to write header.\n");
         ret = 1;
         goto depad_end;
     }
+    if (ga.write_index) {
+        if (!(fn_out_idx = auto_index(out, fn_out, h_fix))) {
+            ret = 1;
+            goto depad_end;
+        }
+    }
 
     // Do the depad
     if (bam_pad2unpad(in, out, h, fai) != 0) ret = 1;
 
+    if (ga.write_index) {
+        if (sam_idx_save(out) < 0) {
+            print_error_errno("depad", "writing index failed");
+            ret = 1;
+        }
+    }
+
 depad_end:
     // close files, free and return
+    free(arg_list);
     if (fai) fai_destroy(fai);
-    if (h) bam_hdr_destroy(h);
+    if (h) sam_hdr_destroy(h);
+    if (h_fix && h_fix != h) sam_hdr_destroy(h_fix);
     if (in) sam_close(in);
     if (out && sam_close(out) < 0) {
         fprintf(samtools_stderr, "[depad] error on closing output file.\n");
         ret = 1;
     }
-    free(fn_list); free(fn_out);
+    free(fn_fai); free(fn_out);
+    if (fn_out_idx)
+        free(fn_out_idx);
+    sam_global_args_free(&ga);
     return ret;
 }
 
@@ -623,8 +608,9 @@ static int usage(int is_long_help)
     fprintf(samtools_stderr, "  -T, --reference FILE\n");
     fprintf(samtools_stderr, "               Padded reference sequence file [null]\n");
     fprintf(samtools_stderr, "  -o FILE      Output file name [samtools_stdout]\n");
+    fprintf(samtools_stderr, "  --no-PG      do not add a PG line\n");
     fprintf(samtools_stderr, "  -?           Longer help\n");
-    sam_global_opt_help(samtools_stderr, "-...--");
+    sam_global_opt_help(samtools_stderr, "-...--..");
 
     if (is_long_help)
         fprintf(samtools_stderr,
