@@ -1,5 +1,4 @@
 import os
-import sys
 import subprocess
 import threading
 import errno
@@ -7,8 +6,6 @@ import unittest
 from pysam import AlignmentFile
 from TestUtils import make_data_files, BAM_DATADIR
 
-IS_PYTHON2 = sys.version_info[0] == 2
-
 
 def setUpModule():
     make_data_files(BAM_DATADIR)
@@ -44,7 +41,6 @@ class StreamTest(unittest.TestCase):
                 read += 1
         return 0, read
 
-    @unittest.skipIf(IS_PYTHON2, "no context manager in py2")
     def test_text_processing(self):
 
         with subprocess.Popen('head -n200',
