@@ -5,7 +5,6 @@ Execute in the :file:`tests` directory as it requires the Makefile
 and data files located there.
 '''
 
-import sys
 import os
 import shutil
 import gzip
@@ -16,8 +15,6 @@ import re
 from TestUtils import checkBinaryEqual, checkGZBinaryEqual, check_url, \
     load_and_convert, make_data_files, TABIX_DATADIR, get_temp_filename
 
-IS_PYTHON3 = sys.version_info[0] >= 3
-
 
 def setUpModule():
     make_data_files(TABIX_DATADIR)
@@ -30,11 +27,8 @@ def myzip_open(infile, mode="r"):
         for l in f:
             yield l.decode("ascii")
 
-    if IS_PYTHON3:
-        if mode == "r":
-            return _convert(gzip.open(infile, "r"))
-    else:
-        return gzip.open(mode)
+    if mode == "r":
+        return _convert(gzip.open(infile, "r"))
 
 
 def splitToBytes(s):
@@ -236,8 +230,7 @@ class TestGZFile(IterationTest):
     with_comments = True
 
     def setUp(self):
-
-        IterationTest.setUp(self)
+        super().setUp()
         self.gzfile = pysam.GZIterator(self.filename)
 
     def testAll(self):
@@ -255,7 +248,7 @@ class TestIterationWithoutComments(IterationTest):
                             "example.gtf.gz")
 
     def setUp(self):
-        IterationTest.setUp(self)
+        super().setUp()
         self.tabix = pysam.TabixFile(self.filename)
 
     def tearDown(self):
@@ -380,9 +373,6 @@ class TestIterationWithComments(TestIterationWithoutComments):
 
     filename = os.path.join(TABIX_DATADIR, "example_comments.gtf.gz")
 
-    def setUp(self):
-        TestIterationWithoutComments.setUp(self)
-
 
 class TestIterators(unittest.TestCase):
     filename = os.path.join(TABIX_DATADIR, "example.gtf.gz")
@@ -511,6 +501,9 @@ class TestBed(unittest.TestCase):
             self.assertEqual(c[0], r.contig)
             self.assertEqual(int(c[1]), r.start)
             self.assertEqual(int(c[2]), r.end)
+            # Needs lambda so that the property getter isn't called too soon
+            self.assertRaises(KeyError, lambda: r.name)
+            self.assertRaises(KeyError, lambda: r.score)
             self.assertEqual(list(c), list(r))
             self.assertEqual("\t".join(map(str, c)),
                              str(r))
@@ -534,6 +527,12 @@ class TestBed(unittest.TestCase):
             self.assertEqual(int(c[2]) + 1, r.end)
             self.assertEqual(str(int(c[2]) + 1), r[2])
 
+            with self.assertRaises(IndexError):
+                r.name = "test"
+
+            with self.assertRaises(IndexError):
+                r.score = 1
+
 
 class TestVCF(unittest.TestCase):
 
@@ -550,46 +549,45 @@ class TestVCF(unittest.TestCase):
             os.unlink(self.tmpfilename + ".gz.tbi")
 
 
-if IS_PYTHON3:
-    class TestUnicode(unittest.TestCase):
+class TestUnicode(unittest.TestCase):
 
-        '''test reading from a file with non-ascii characters.'''
+    '''test reading from a file with non-ascii characters.'''
 
-        filename = os.path.join(TABIX_DATADIR, "example_unicode.vcf")
+    filename = os.path.join(TABIX_DATADIR, "example_unicode.vcf")
 
-        def setUp(self):
-            self.tmpfilename = get_temp_filename(suffix="vcf")
-            shutil.copyfile(self.filename, self.tmpfilename)
-            pysam.tabix_index(self.tmpfilename, preset="vcf")
+    def setUp(self):
+        self.tmpfilename = get_temp_filename(suffix="vcf")
+        shutil.copyfile(self.filename, self.tmpfilename)
+        pysam.tabix_index(self.tmpfilename, preset="vcf")
 
-        def tearDown(self):
-            os.unlink(self.tmpfilename + ".gz")
-            if os.path.exists(self.tmpfilename + ".gz.tbi"):
-                os.unlink(self.tmpfilename + ".gz.tbi")
+    def tearDown(self):
+        os.unlink(self.tmpfilename + ".gz")
+        if os.path.exists(self.tmpfilename + ".gz.tbi"):
+            os.unlink(self.tmpfilename + ".gz.tbi")
 
-        def testFromTabix(self):
+    def testFromTabix(self):
 
-            # use ascii encoding - should raise error
-            with pysam.TabixFile(
-                    self.tmpfilename + ".gz", encoding="ascii") as t:
-                results = list(t.fetch(parser=pysam.asVCF()))
-                self.assertRaises(UnicodeDecodeError,
-                                  getattr, results[1], "id")
+        # use ascii encoding - should raise error
+        with pysam.TabixFile(
+                self.tmpfilename + ".gz", encoding="ascii") as t:
+            results = list(t.fetch(parser=pysam.asVCF()))
+            self.assertRaises(UnicodeDecodeError,
+                              getattr, results[1], "id")
 
-            with pysam.TabixFile(
-                    self.tmpfilename + ".gz", encoding="utf-8") as t:
-                results = list(t.fetch(parser=pysam.asVCF()))
-                self.assertEqual(getattr(results[1], "id"), u"Rene\xe9")
+        with pysam.TabixFile(
+                self.tmpfilename + ".gz", encoding="utf-8") as t:
+            results = list(t.fetch(parser=pysam.asVCF()))
+            self.assertEqual(getattr(results[1], "id"), u"Rene\xe9")
 
-        def testFromVCF(self):
-            self.vcf = pysam.VCF()
-            self.assertRaises(
-                UnicodeDecodeError,
-                self.vcf.connect,
-                self.tmpfilename + ".gz",
-                "ascii")
-            self.vcf.connect(self.tmpfilename + ".gz", encoding="utf-8")
-            v = self.vcf.getsamples()[0]
+    def testFromVCF(self):
+        self.vcf = pysam.VCF()
+        self.assertRaises(
+            UnicodeDecodeError,
+            self.vcf.connect,
+            self.tmpfilename + ".gz",
+            "ascii")
+        self.vcf.connect(self.tmpfilename + ".gz", encoding="utf-8")
+        v = self.vcf.getsamples()[0]
 
 
 class TestVCFFromTabix(TestVCF):
@@ -599,15 +597,14 @@ class TestVCFFromTabix(TestVCF):
                "filter", "info", "format")
 
     def setUp(self):
-
-        TestVCF.setUp(self)
+        super().setUp()
 
         self.tabix = pysam.TabixFile(self.tmpfilename + ".gz")
         self.compare = load_and_convert(self.filename)
 
     def tearDown(self):
         self.tabix.close()
-        TestVCF.tearDown(self)
+        super().tearDown()
 
     def testRead(self):
 
@@ -708,14 +705,13 @@ class TestVCFFromVCF(TestVCF):
     missing_quality = -1
 
     def setUp(self):
-
-        TestVCF.setUp(self)
+        super().setUp()
 
         self.vcf = pysam.VCF()
         self.compare = load_and_convert(self.filename, encode=False)
 
     def tearDown(self):
-        TestVCF.tearDown(self)
+        super().tearDown()
         self.vcf.close()
 
     def open_vcf(self, fn):
